def iterate_node_executions(
    client,
    workflow_execution_identifier=None,
    task_execution_identifier=None,
    limit=None,
    filters=None,
):
    """
    This returns a generator for node executions.
    :param flytekit.clients.friendly.SynchronousFlyteClient client:
    :param flytekit.common.core.identifier.WorkflowExecutionIdentifier workflow_execution_identifier:
    :param flytekit.common.core.identifier.TaskExecutionIdentifier task_execution_identifier:
    :param int limit: The maximum number of elements to retrieve
    :param list[flytekit.models.filters.Filter] filters:
    :rtype: Iterator[flytekit.models.node_execution.NodeExecution]
    """
    token = ""
    num_to_fetch = 100
    if limit is not None and limit < num_to_fetch:
        num_to_fetch = limit
    counter = 0
    while True:
        if workflow_execution_identifier is not None:
            node_execs, next_token = client.list_node_executions(
                workflow_execution_identifier=workflow_execution_identifier,
                limit=num_to_fetch,
                token=token,
                filters=filters,
            )
        else:
            node_execs, next_token = client.list_node_executions_for_task_paginated(
                task_execution_identifier=task_execution_identifier,
                limit=num_to_fetch,
                token=token,
                filters=filters,
            )
        for n in node_execs:
            counter += 1
            if limit is not None and counter > limit:
                return
            yield n
        if not next_token:
            break
        token = next_token


def iterate_task_executions(client, node_execution_identifier, limit=None, filters=None):
    """
    This returns a generator for task executions, given a node execution identifier
    :param flytekit.clients.friendly.SynchronousFlyteClient client:
    :param flytekit.models.core.identifier.NodeExecutionIdentifier node_execution_identifier:
    :param int limit: The maximum number of elements to retrieve
    :param list[flytekit.models.filters.Filter] filters:
    :rtype: Iterator[flytekit.models.admin.task_execution.TaskExecution]
    """
    token = ""
    num_to_fetch = 100
    if limit is not None and limit < num_to_fetch:
        num_to_fetch = limit
    counter = 0
    while True:
        task_execs, next_token = client.list_task_executions_paginated(
            node_execution_identifier=node_execution_identifier,
            limit=num_to_fetch,
            token=token,
            filters=filters,
        )
        for t in task_execs:
            counter += 1
            if limit is not None and counter > limit:
                return
            yield t
        if not next_token:
            break
        token = next_token



import datetime
import os
import pathlib
from typing import Optional, Union

from flytekit.common import constants as _constants
from flytekit.common import utils as _common_utils
from flytekit.common.exceptions import user as _user_exception
from flytekit.configuration import platform as _platform_config
from flytekit.configuration import sdk as _sdk_config
from flytekit.interfaces.data.gcs import gcs_proxy as _gcs_proxy
from flytekit.interfaces.data.http import http_data_proxy as _http_data_proxy
from flytekit.interfaces.data.local import local_file_proxy as _local_file_proxy
from flytekit.interfaces.data.s3 import s3proxy as _s3proxy
from flytekit.loggers import logger


class LocalWorkingDirectoryContext(object):
    _CONTEXTS = []

    def __init__(self, directory):
        self._directory = directory

    def __enter__(self):
        self._CONTEXTS.append(self._directory)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._CONTEXTS.pop()

    @classmethod
    def get(cls):
        return cls._CONTEXTS[-1] if cls._CONTEXTS else None


class _OutputDataContext(object):
    _CONTEXTS = [_local_file_proxy.LocalFileProxy(_sdk_config.LOCAL_SANDBOX.get())]

    def __init__(self, context):
        self._context = context

    def __enter__(self):
        self._CONTEXTS.append(self._context)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._CONTEXTS.pop()

    @classmethod
    def get_active_proxy(cls):
        return cls._CONTEXTS[-1]

    @classmethod
    def get_default_proxy(cls):
        return cls._CONTEXTS[0]


class LocalDataContext(_OutputDataContext):
    def __init__(self, sandbox):
        """
        :param Text sandbox:
        """
        super(LocalDataContext, self).__init__(_local_file_proxy.LocalFileProxy(sandbox))


class RemoteDataContext(_OutputDataContext):
    _CLOUD_PROVIDER_TO_PROXIES = {
        _constants.CloudProvider.AWS: _s3proxy.AwsS3Proxy,
        _constants.CloudProvider.GCP: _gcs_proxy.GCSProxy,
    }

    def __init__(self, cloud_provider=None, raw_output_data_prefix_override=None):
        """
        :param Optional[Text] cloud_provider: From flytekit.common.constants.CloudProvider enum
        """
        cloud_provider = cloud_provider or _platform_config.CLOUD_PROVIDER.get()
        proxy_class = type(self)._CLOUD_PROVIDER_TO_PROXIES.get(cloud_provider, None)
        if proxy_class is None:
            raise _user_exception.FlyteAssertion(
                "Configured cloud provider is not supported for data I/O.  Received: {}, expected one of: {}".format(
                    cloud_provider, list(type(self)._CLOUD_PROVIDER_TO_PROXIES.keys())
                )
            )
        proxy = proxy_class(raw_output_data_prefix_override)
        super(RemoteDataContext, self).__init__(proxy)


class Data(object):
    # TODO: More proxies for more environments.
    _DATA_PROXIES = {
        "s3:/": _s3proxy.AwsS3Proxy(),
        "gs:/": _gcs_proxy.GCSProxy(),
        "http://": _http_data_proxy.HttpFileProxy(),
        "https://": _http_data_proxy.HttpFileProxy(),
    }

    @classmethod
    def _load_data_proxy_by_path(cls, path):
        """
        :param Text path:
        :rtype: flytekit.interfaces.data.common.DataProxy
        """
        for k, v in cls._DATA_PROXIES.items():
            if path.startswith(k):
                return v
        return _OutputDataContext.get_default_proxy()

    @classmethod
    def data_exists(cls, path):
        """
        :param Text path:
        :rtype: bool: whether the file exists or not
        """
        with _common_utils.PerformanceTimer("Check file exists {}".format(path)):
            proxy = cls._load_data_proxy_by_path(path)
            return proxy.exists(path)

    @classmethod
    def get_data(cls, remote_path, local_path, is_multipart=False):
        """
        :param Text remote_path:
        :param Text local_path:
        :param bool is_multipart:
        """
        try:
            with _common_utils.PerformanceTimer("Copying ({} -> {})".format(remote_path, local_path)):
                proxy = cls._load_data_proxy_by_path(remote_path)
                if is_multipart:
                    proxy.download_directory(remote_path, local_path)
                else:
                    proxy.download(remote_path, local_path)
        except Exception as ex:
            raise _user_exception.FlyteAssertion(
                "Failed to get data from {remote_path} to {local_path} (recursive={is_multipart}).\n\n"
                "Original exception: {error_string}".format(
                    remote_path=remote_path,
                    local_path=local_path,
                    is_multipart=is_multipart,
                    error_string=str(ex),
                )
            )

    @classmethod
    def put_data(cls, local_path, remote_path, is_multipart=False):
        """
        :param Text local_path:
        :param Text remote_path:
        :param bool is_multipart:
        """
        try:
            with _common_utils.PerformanceTimer("Writing ({} -> {})".format(local_path, remote_path)):
                proxy = cls._load_data_proxy_by_path(remote_path)
                if is_multipart:
                    proxy.upload_directory(local_path, remote_path)
                else:
                    proxy.upload(local_path, remote_path)
        except Exception as ex:
            raise _user_exception.FlyteAssertion(
                "Failed to put data from {local_path} to {remote_path} (recursive={is_multipart}).\n\n"
                "Original exception: {error_string}".format(
                    remote_path=remote_path,
                    local_path=local_path,
                    is_multipart=is_multipart,
                    error_string=str(ex),
                )
            )

    @classmethod
    def get_remote_path(cls):
        """
        :rtype: Text
        """
        return _OutputDataContext.get_active_proxy().get_random_path()

    @classmethod
    def get_remote_directory(cls):
        """
        :rtype: Text
        """
        return _OutputDataContext.get_active_proxy().get_random_directory()


class FileAccessProvider(object):
    def __init__(
        self,
        local_sandbox_dir: Union[str, os.PathLike],
        remote_proxy: Union[_s3proxy.AwsS3Proxy, _gcs_proxy.GCSProxy, None] = None,
    ):

        # Local access
        if local_sandbox_dir is None or local_sandbox_dir == "":
            raise Exception("Can't use empty path")
        local_sandbox_dir_appended = os.path.join(local_sandbox_dir, "local_flytekit")
        pathlib.Path(local_sandbox_dir_appended).mkdir(parents=True, exist_ok=True)
        self._local_sandbox_dir = local_sandbox_dir_appended
        self._local = _local_file_proxy.LocalFileProxy(local_sandbox_dir_appended)

        # Remote/cloud stuff
        if isinstance(remote_proxy, _s3proxy.AwsS3Proxy):
            self._aws = remote_proxy
        if isinstance(remote_proxy, _gcs_proxy.GCSProxy):
            self._gcs = remote_proxy
        if remote_proxy is not None:
            self._remote = remote_proxy
        else:
            mock_remote = os.path.join(local_sandbox_dir, "mock_remote")
            pathlib.Path(mock_remote).mkdir(parents=True, exist_ok=True)
            self._remote = _local_file_proxy.LocalFileProxy(mock_remote)

        # HTTP access
        self._http_proxy = _http_data_proxy.HttpFileProxy()

    @staticmethod
    def is_remote(path: Union[str, os.PathLike]) -> bool:
        if path.startswith("s3:/") or path.startswith("gs:/") or path.startswith("file:/") or path.startswith("http"):
            return True
        return False

    def _get_data_proxy_by_path(self, path: Union[str, os.PathLike]):
        """
        :param Text path:
        :rtype: flytekit.interfaces.data.common.DataProxy
        """
        if path.startswith("s3:/"):
            return self.aws
        elif path.startswith("gs:/"):
            return self.gcs
        elif path.startswith("http"):
            return self.http
        elif path.startswith("file://"):
            # Note that we default to the local one here, not the remote one.
            return self.local_access
        elif path.startswith("/"):
            # Note that we default to the local one here, not the remote one.
            return self.local_access
        raise Exception(f"Unknown file access {path}")

    @property
    def aws(self) -> _s3proxy.AwsS3Proxy:
        if self._aws is None:
            raise Exception("No AWS handler found")
        return self._aws

    @property
    def gcs(self) -> _gcs_proxy.GCSProxy:
        if self._gcs is None:
            raise Exception("No GCP handler found")
        return self._gcs

    @property
    def remote(self):
        if self._remote is not None:
            return self._remote
        raise Exception("No cloud provider specified")

    @property
    def http(self) -> _http_data_proxy.HttpFileProxy:
        return self._http_proxy

    @property
    def local_sandbox_dir(self) -> os.PathLike:
        return self._local_sandbox_dir

    @property
    def local_access(self) -> _local_file_proxy.LocalFileProxy:
        return self._local

    def get_random_remote_path(self, file_path_or_file_name: Optional[str] = None) -> str:
        """
        :param file_path_or_file_name: For when you want a random directory, but want to preserve the leaf file name
        """
        if file_path_or_file_name:
            _, tail = os.path.split(file_path_or_file_name)
            if tail:
                return f"{self.remote.get_random_directory()}{tail}"
            else:
                logger.warning(f"No filename detected in {file_path_or_file_name}, using random remote path...")
        return self.remote.get_random_path()

    def get_random_remote_directory(self):
        return self.remote.get_random_directory()

    def get_random_local_path(self, file_path_or_file_name: Optional[str] = None) -> str:
        """
        :param file_path_or_file_name:  For when you want a random directory, but want to preserve the leaf file name
        """
        if file_path_or_file_name:
            _, tail = os.path.split(file_path_or_file_name)
            if tail:
                return os.path.join(self.local_access.get_random_directory(), tail)
            else:
                logger.warning(f"No filename detected in {file_path_or_file_name}, using random local path...")
        return self.local_access.get_random_path()

    def get_random_local_directory(self) -> str:
        dir = self.local_access.get_random_directory()
        pathlib.Path(dir).mkdir(parents=True, exist_ok=True)
        return dir

    def exists(self, remote_path: str) -> bool:
        """
        :param Text remote_path: remote s3:// or gs:// path
        """
        return self._get_data_proxy_by_path(remote_path).exists(remote_path)

    def download_directory(self, remote_path: str, local_path: str):
        """
        :param Text remote_path: remote s3:// path
        :param Text local_path: directory to copy to
        """
        return self._get_data_proxy_by_path(remote_path).download_directory(remote_path, local_path)

    def download(self, remote_path: str, local_path: str):
        """
        :param Text remote_path: remote s3:// path
        :param Text local_path: directory to copy to
        """
        return self._get_data_proxy_by_path(remote_path).download(remote_path, local_path)

    def upload(self, file_path: str, to_path: str):
        """
        :param Text file_path:
        :param Text to_path:
        """
        return self.remote.upload(file_path, to_path)

    def upload_directory(self, local_path: str, remote_path: str):
        """
        :param Text local_path:
        :param Text remote_path:
        """
        # TODO: Clean this up, this is a minor hack in lieu of https://github.com/flyteorg/flyte/issues/762
        if remote_path.startswith("/"):
            return self.local_access.upload_directory(local_path, remote_path)
        return self.remote.upload_directory(local_path, remote_path)

    def get_data(self, remote_path: str, local_path: str, is_multipart=False):
        """
        :param Text remote_path:
        :param Text local_path:
        :param bool is_multipart:
        """
        try:
            with _common_utils.PerformanceTimer("Copying ({} -> {})".format(remote_path, local_path)):
                if is_multipart:
                    self.download_directory(remote_path, local_path)
                else:
                    self.download(remote_path, local_path)
        except Exception as ex:
            raise _user_exception.FlyteAssertion(
                "Failed to get data from {remote_path} to {local_path} (recursive={is_multipart}).\n\n"
                "Original exception: {error_string}".format(
                    remote_path=remote_path,
                    local_path=local_path,
                    is_multipart=is_multipart,
                    error_string=str(ex),
                )
            )

    def put_data(self, local_path: Union[str, os.PathLike], remote_path: str, is_multipart=False):
        """
        The implication here is that we're always going to put data to the remote location, so we .remote to ensure
        we don't use the true local proxy if the remote path is a file://

        :param Text local_path:
        :param Text remote_path:
        :param bool is_multipart:
        """
        try:
            with _common_utils.PerformanceTimer("Writing ({} -> {})".format(local_path, remote_path)):
                if is_multipart:
                    self.remote.upload_directory(local_path, remote_path)
                else:
                    self.remote.upload(local_path, remote_path)
        except Exception as ex:
            raise _user_exception.FlyteAssertion(
                f"Failed to put data from {local_path} to {remote_path} (recursive={is_multipart}).\n\n"
                f"Original exception: {str(ex)}"
            ) from ex


timestamped_default_sandbox_location = os.path.join(
    _sdk_config.LOCAL_SANDBOX.get(), datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
)
default_local_file_access_provider = FileAccessProvider(local_sandbox_dir=timestamped_default_sandbox_location)



from flytekit.configuration import common as _config_common

ASSUMABLE_IAM_ROLE = _config_common.FlyteStringConfigurationEntry("auth", "assumable_iam_role", default=None)
"""
This is the role the SDK will use by default to execute workflows.  For example, in AWS this should be an IAM role
string.
"""

KUBERNETES_SERVICE_ACCOUNT = _config_common.FlyteStringConfigurationEntry(
    "auth", "kubernetes_service_account", default=None
)
"""
This is the kubernetes service account that will be passed to workflow executions.
"""

RAW_OUTPUT_DATA_PREFIX = _config_common.FlyteStringConfigurationEntry("auth", "raw_output_data_prefix", default="")
"""
This is not output metadata but rather where users can specify an S3 or gcs path for offloaded data like blobs
and schemas.

The reason this setting is in this file is because it's inextricably tied to a workflow's role or service account,
since that is what ultimately gives the tasks the ability to write to certain buckets.
"""



"""
=====================
Extending Flytekit
=====================

.. currentmodule:: flytekit.extend

This package contains things that are useful when extending Flytekit.

.. autosummary::
   :toctree: generated/

   get_serializable
   context_manager
   SQLTask
   IgnoreOutputs
   PythonTask
   ExecutionState
   Image
   ImageConfig
   SerializationSettings
   Interface
   Promise
   TaskPlugins
   DictTransformer
   T
   TypeEngine
   TypeTransformer

"""

from flytekit.common.translator import get_serializable
from flytekit.core import context_manager
from flytekit.core.base_sql_task import SQLTask
from flytekit.core.base_task import IgnoreOutputs, PythonTask, TaskResolverMixin
from flytekit.core.class_based_resolver import ClassStorageTaskResolver
from flytekit.core.context_manager import ExecutionState, Image, ImageConfig, SerializationSettings
from flytekit.core.interface import Interface
from flytekit.core.promise import Promise
from flytekit.core.task import TaskPlugins
from flytekit.core.type_engine import DictTransformer, T, TypeEngine, TypeTransformer



from typing import List

from flytekit.core.base_task import TaskResolverMixin
from flytekit.core.context_manager import SerializationSettings
from flytekit.core.python_auto_container import PythonAutoContainerTask
from flytekit.core.tracker import TrackedInstance


class ClassStorageTaskResolver(TrackedInstance, TaskResolverMixin):
    """
    Stores tasks inside a class variable. The class must be inherited from at the point of usage because the task
    loading process basically relies on the same sequence of things happening.
    """

    def __init__(self, *args, **kwargs):
        self.mapping = []
        super().__init__(*args, **kwargs)

    def name(self) -> str:
        return "ClassStorageTaskResolver"

    def get_all_tasks(self) -> List[PythonAutoContainerTask]:
        return self.mapping

    def add(self, t: PythonAutoContainerTask):
        self.mapping.append(t)

    def load_task(self, loader_args: List[str]) -> PythonAutoContainerTask:
        if len(loader_args) != 1:
            raise RuntimeError(f"Unable to load task, received ambiguous loader args {loader_args}, expected only one")

        # string should be parseable as an int
        idx = int(loader_args[0])
        return self.mapping[idx]

    def loader_args(self, settings: SerializationSettings, t: PythonAutoContainerTask) -> List[str]:
        """
        This is responsible for turning an instance of a task into args that the load_task function can reconstitute.
        """
        if t not in self.mapping:
            raise Exception("no such task")

        return [f"{self.mapping.index(t)}"]



import pandas
import pyspark
from flytekitplugins.spark.task import Spark

import flytekit
from flytekit import kwtypes, task, workflow
from flytekit.types.schema import FlyteSchema


def test_wf1_with_spark():
    @task(task_config=Spark())
    def my_spark(a: int) -> (int, str):
        session = flytekit.current_context().spark_session
        assert session.sparkContext.appName == "FlyteSpark: ex:local:local:local"
        return a + 2, "world"

    @task
    def t2(a: str, b: str) -> str:
        return b + a

    @workflow
    def my_wf(a: int, b: str) -> (int, str):
        x, y = my_spark(a=a)
        d = t2(a=y, b=b)
        return x, d

    x = my_wf(a=5, b="hello ")
    assert x == (7, "hello world")


def test_spark_dataframe_input():
    my_schema = FlyteSchema[kwtypes(name=str, age=int)]

    @task
    def my_dataset() -> my_schema:
        return pandas.DataFrame(data={"name": ["Alice"], "age": [5]})

    @task(task_config=Spark())
    def my_spark(df: pyspark.sql.DataFrame) -> my_schema:
        session = flytekit.current_context().spark_session
        new_df = session.createDataFrame([("Bob", 10)], my_schema.column_names())
        return df.union(new_df)

    @workflow
    def my_wf() -> my_schema:
        df = my_dataset()
        return my_spark(df=df)

    x = my_wf()
    assert x
    reader = x.open()
    df2 = reader.all()
    assert df2 is not None


def test_spark_dataframe_return():
    my_schema = FlyteSchema[kwtypes(name=str, age=int)]

    @task(task_config=Spark())
    def my_spark(a: int) -> my_schema:
        session = flytekit.current_context().spark_session
        df = session.createDataFrame([("Alice", a)], my_schema.column_names())
        print(type(df))
        return df

    @workflow
    def my_wf(a: int) -> my_schema:
        return my_spark(a=a)

    x = my_wf(a=5)
    reader = x.open(pandas.DataFrame)
    df2 = reader.all()
    result_df = df2.reset_index(drop=True) == pandas.DataFrame(data={"name": ["Alice"], "age": [5]}).reset_index(
        drop=True
    )
    assert result_df.all().all()



import hashlib as _hashlib
import json as _json
import logging as _logging
import uuid as _uuid

import six as _six
from google.protobuf import json_format as _json_format
from google.protobuf import struct_pb2 as _struct

from flytekit.common import interface as _interfaces
from flytekit.common import nodes as _nodes
from flytekit.common import sdk_bases as _sdk_bases
from flytekit.common import workflow_execution as _workflow_execution
from flytekit.common.core import identifier as _identifier
from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.mixins import hash as _hash_mixin
from flytekit.common.mixins import launchable as _launchable_mixin
from flytekit.common.mixins import registerable as _registerable
from flytekit.common.types import helpers as _type_helpers
from flytekit.configuration import auth as _auth_config
from flytekit.configuration import internal as _internal_config
from flytekit.configuration import sdk as _sdk_config
from flytekit.engines.flyte import engine as _flyte_engine
from flytekit.models import common as _common_model
from flytekit.models import execution as _admin_execution_models
from flytekit.models import task as _task_model
from flytekit.models.admin import common as _admin_common
from flytekit.models.core import identifier as _identifier_model
from flytekit.models.core import workflow as _workflow_model


class SdkTask(
    _hash_mixin.HashOnReferenceMixin,
    _registerable.RegisterableEntity,
    _launchable_mixin.LaunchableEntity,
    _task_model.TaskTemplate,
    metaclass=_sdk_bases.ExtendedSdkType,
):
    def __init__(
        self, type, metadata, interface, custom, container=None, task_type_version=0, security_context=None, config=None
    ):
        """
        :param Text type: This is used to define additional extensions for use by Propeller or SDK.
        :param TaskMetadata metadata: This contains information needed at runtime to determine behavior such as
            whether or not outputs are discoverable, timeouts, and retries.
        :param flytekit.common.interface.TypedInterface interface: The interface definition for this task.
        :param dict[Text, T] custom: Arbitrary type for use by plugins.
        :param Container container: Provides the necessary entrypoint information for execution.  For instance,
            a Container might be specified with the necessary command line arguments.
        :param int task_type_version: Specific version of this task type used by plugins to potentially modify
            execution behavior or serialization.
        :param _SecurityContext security_context:
        """
        # TODO: Remove the identifier portion and fill in with local values.
        super(SdkTask, self).__init__(
            _identifier.Identifier(
                _identifier_model.ResourceType.TASK,
                _internal_config.PROJECT.get(),
                _internal_config.DOMAIN.get(),
                _uuid.uuid4().hex,
                _internal_config.VERSION.get(),
            ),
            type,
            metadata,
            interface,
            custom,
            container=container,
            task_type_version=task_type_version,
            security_context=security_context,
            config=config,
        )

    @property
    def interface(self):
        """
        :rtype: flytekit.common.interface.TypedInterface
        """
        return super(SdkTask, self).interface

    @property
    def resource_type(self):
        """
        Integer from _identifier.ResourceType enum
        :rtype: int
        """
        return _identifier_model.ResourceType.TASK

    @property
    def entity_type_text(self):
        """
        :rtype: Text
        """
        return "Task"

    @classmethod
    def promote_from_model(cls, base_model):
        """
        :param flytekit.models.task.TaskTemplate base_model:
        :rtype: SdkTask
        """
        t = cls(
            type=base_model.type,
            metadata=base_model.metadata,
            interface=_interfaces.TypedInterface.promote_from_model(base_model.interface),
            custom=base_model.custom,
            container=base_model.container,
            task_type_version=base_model.task_type_version,
        )
        # Override the newly generated name if one exists in the base model
        if not base_model.id.is_empty:
            t._id = _identifier.Identifier.promote_from_model(base_model.id)

        return t

    def assign_custom_and_return(self, custom):
        self._custom = custom
        return self

    def assign_type_and_return(self, new_type):
        self._type = new_type
        return self

    @_exception_scopes.system_entry_point
    def __call__(self, *args, **input_map):
        """
        :param list[T] args: Do not specify.  Kwargs only are supported for this function.
        :param dict[str, T] input_map: Map of inputs.  Can be statically defined or OutputReference links.
        :rtype: flytekit.common.nodes.SdkNode
        """
        if len(args) > 0:
            raise _user_exceptions.FlyteAssertion(
                "When adding a task as a node in a workflow, all inputs must be specified with kwargs only.  We "
                "detected {} positional args.".format(len(args))
            )

        bindings, upstream_nodes = self.interface.create_bindings_for_inputs(input_map)

        # TODO: Remove DEADBEEF
        # One thing to note - this function is not overloaded at the SdkRunnableTask layer, which means 'self' here
        # will sometimes refer to an object that can be executed locally, and other times will refer to something
        # that cannot (ie a pure SdkTask object, fetched from Admin for instance).
        return _nodes.SdkNode(
            id=None,
            metadata=_workflow_model.NodeMetadata(
                "DEADBEEF",
                self.metadata.timeout,
                self.metadata.retries,
                self.metadata.interruptible,
            ),
            bindings=sorted(bindings, key=lambda b: b.var),
            upstream_nodes=upstream_nodes,
            sdk_task=self,
        )

    @_exception_scopes.system_entry_point
    def register(self, project, domain, name, version):
        """
        :param Text project: The project in which to register this task.
        :param Text domain: The domain in which to register this task.
        :param Text name: The name to give this task.
        :param Text version: The version in which to register this task.
        """
        # TODO: Revisit the notion of supplying the project, domain, name, version, as opposed to relying on the
        #       current ID.
        self.validate()
        id_to_register = _identifier.Identifier(_identifier_model.ResourceType.TASK, project, domain, name, version)
        old_id = self.id

        client = _flyte_engine.get_client()
        try:
            self._id = id_to_register
            client.create_task(id_to_register, _task_model.TaskSpec(self))
            self._id = old_id
            self._has_registered = True
            return str(id_to_register)
        except _user_exceptions.FlyteEntityAlreadyExistsException:
            pass
        except Exception:
            self._id = old_id
            raise

    @_exception_scopes.system_entry_point
    def serialize(self):
        """
        :rtype: flyteidl.admin.task_pb2.TaskSpec
        """
        return _task_model.TaskSpec(self).to_flyte_idl()

    @classmethod
    @_exception_scopes.system_entry_point
    def fetch(cls, project, domain, name, version):
        """
        This function uses the engine loader to call create a hydrated task from Admin.
        :param Text project:
        :param Text domain:
        :param Text name:
        :param Text version:
        :rtype: SdkTask
        """
        task_id = _identifier.Identifier(_identifier_model.ResourceType.TASK, project, domain, name, version)
        admin_task = _flyte_engine.get_client().get_task(task_id)

        sdk_task = cls.promote_from_model(admin_task.closure.compiled_task.template)
        sdk_task._id = task_id
        sdk_task._has_registered = True
        return sdk_task

    @classmethod
    @_exception_scopes.system_entry_point
    def fetch_latest(cls, project, domain, name):
        """
        This function uses the engine loader to call create a latest hydrated task from Admin.
        :param Text project:
        :param Text domain:
        :param Text name:
        :rtype: SdkTask
        """
        named_task = _common_model.NamedEntityIdentifier(project, domain, name)
        client = _flyte_engine.get_client()
        task_list, _ = client.list_tasks_paginated(
            named_task,
            limit=1,
            sort_by=_admin_common.Sort("created_at", _admin_common.Sort.Direction.DESCENDING),
        )
        admin_task = task_list[0] if task_list else None

        if not admin_task:
            raise _user_exceptions.FlyteEntityNotExistException("Named task {} not found".format(named_task))
        sdk_task = cls.promote_from_model(admin_task.closure.compiled_task.template)
        sdk_task._id = admin_task.id
        return sdk_task

    @_exception_scopes.system_entry_point
    def validate(self):
        pass

    @_exception_scopes.system_entry_point
    def add_inputs(self, inputs):
        raise _user_exceptions.FlyteUserException("You can not add inputs to this task")

    @_exception_scopes.system_entry_point
    def add_outputs(self, outputs):
        """
        Adds the outputs to this task.  This can be called multiple times, but it will fail if an output with a given
        name is added more than once, a name collides with an input, or if the name doesn't exist as an arg name in
        the wrapped function.
        :param dict[Text, flytekit.models.interface.Variable] outputs: names and variables to add as outputs
            to this task
        """
        self._validate_outputs(outputs)
        self.interface.outputs.update(outputs)

    def _validate_inputs(self, inputs):
        """
        This method should be overridden in sub-classes that intend to do additional checks on inputs.  If validation
        fails, this function should raise an informative exception.
        :param dict[Text, flytekit.models.interface.Variable] inputs:  Input variables to validate
        :raises: flytekit.common.exceptions.user.FlyteValidationException
        """
        for k, v in _six.iteritems(inputs):
            if k in self.interface.inputs:
                raise _user_exceptions.FlyteValidationException(
                    "An input with name '{}' is already defined.  Redefinition is not allowed.".format(k)
                )
            if k in self.interface.outputs:
                raise _user_exceptions.FlyteValidationException(
                    "An output with name '{}' is already defined.  Therefore '{}' can't be defined as an "
                    "input".format(k, v)
                )

    def _validate_outputs(self, outputs):
        """
        This method should be overridden in sub-classes that intend to do additional checks on outputs.  If validation
        fails, this function should raise an informative exception.
        :param dict[Text, flytekit.models.interface.Variable] outputs:  Output variables to validate
        :raises: flytekit.common.exceptions.user.FlyteValidationException
        """
        for k, v in _six.iteritems(outputs):
            if k in self.interface.outputs:
                raise _user_exceptions.FlyteValidationException(
                    "An output with name '{}' is already defined.  Redefinition is not allowed.".format(k)
                )
            if k in self.interface.inputs:
                raise _user_exceptions.FlyteValidationException(
                    "An input with name '{}' is already defined.  Therefore '{}' can't be defined as an "
                    "input".format(k, v)
                )

    def __repr__(self):
        return "Flyte {task_type}: {interface}".format(task_type=self.type, interface=self.interface)

    def _python_std_input_map_to_literal_map(self, inputs):
        """
        :param dict[Text,Any] inputs: A dictionary of Python standard inputs that will be type-checked and compiled
            to a LiteralMap
        :rtype: flytekit.models.literals.LiteralMap
        """
        return _type_helpers.pack_python_std_map_to_literal_map(
            inputs,
            {k: _type_helpers.get_sdk_type_from_literal_type(v.type) for k, v in _six.iteritems(self.interface.inputs)},
        )

    def _produce_deterministic_version(self, version=None):
        """
        :param Text version:
        :return Text:
        """

        if self.container is not None and self.container.data_loading_config is None:
            # Only in the case of raw container tasks (which are the only valid tasks with container definitions that
            # can assign a client-side task version) their data config will be None.
            raise ValueError("Client-side task versions are not supported for {} task type".format(self.type))
        if version is not None:
            return version
        custom = _json_format.Parse(_json.dumps(self.custom, sort_keys=True), _struct.Struct()) if self.custom else None

        # The task body is the entirety of the task template MINUS the identifier. The identifier is omitted because
        # 1) this method is used to compute the version portion of the identifier and
        # 2 ) the SDK will actually generate a unique name on every task instantiation which is not great for
        # the reproducibility this method attempts.
        task_body = (
            self.type,
            self.metadata.to_flyte_idl().SerializeToString(deterministic=True),
            self.interface.to_flyte_idl().SerializeToString(deterministic=True),
            custom,
        )
        return _hashlib.md5(str(task_body).encode("utf-8")).hexdigest()

    @_exception_scopes.system_entry_point
    def register_and_launch(self, project, domain, name, version=None, inputs=None):
        """
        :param Text project: The project in which to register and launch this task.
        :param Text domain: The domain in which to register and launch this task.
        :param Text name: The name to give this task.
        :param Text version: The version in which to register this task
        :param dict[Text, Any] inputs: A dictionary of Python standard inputs that will be type-checked, then compiled
            to a LiteralMap.

        :rtype: flytekit.common.workflow_execution.SdkWorkflowExecution
        """
        self.validate()
        version = self._produce_deterministic_version(version)
        self.register(project, domain, name, version)
        return self.launch(project, domain, inputs=inputs)

    @_exception_scopes.system_entry_point
    def launch_with_literals(
        self,
        project,
        domain,
        literal_inputs,
        name=None,
        notification_overrides=None,
        label_overrides=None,
        annotation_overrides=None,
        auth_role=None,
    ):
        """
        Launches a single task execution and returns the execution identifier.
        :param Text project:
        :param Text domain:
        :param flytekit.models.literals.LiteralMap literal_inputs: Inputs to the execution.
        :param Text name: [Optional] If specified, an execution will be created with this name.  Note: the name must
            be unique within the context of the project and domain.
        :param list[flytekit.common.notifications.Notification] notification_overrides: [Optional] If specified, these
            are the notifications that will be honored for this execution.  An empty list signals to disable all
            notifications.
        :param flytekit.models.common.Labels label_overrides:
        :param flytekit.models.common.Annotations annotation_overrides:
        :param flytekit.models.common.AuthRole auth_role:
        :rtype: flytekit.common.workflow_execution.SdkWorkflowExecution
        """
        disable_all = notification_overrides == []
        if disable_all:
            notification_overrides = None
        else:
            notification_overrides = _admin_execution_models.NotificationList(notification_overrides or [])
            disable_all = None

        # Unlike regular workflow executions, single task executions must always specify an auth role, since there isn't
        # any existing launch plan with a bound auth role to fall back on.
        if auth_role is None:
            assumable_iam_role = _auth_config.ASSUMABLE_IAM_ROLE.get()
            kubernetes_service_account = _auth_config.KUBERNETES_SERVICE_ACCOUNT.get()

            if not (assumable_iam_role or kubernetes_service_account):
                _logging.warning(
                    "Using deprecated `role` from config. "
                    "Please update your config to use `assumable_iam_role` instead"
                )
                assumable_iam_role = _sdk_config.ROLE.get()
            auth_role = _common_model.AuthRole(
                assumable_iam_role=assumable_iam_role,
                kubernetes_service_account=kubernetes_service_account,
            )

        client = _flyte_engine.get_client()
        try:
            # TODO(katrogan): Add handling to register the underlying task if it's not already.
            exec_id = client.create_execution(
                project,
                domain,
                name,
                _admin_execution_models.ExecutionSpec(
                    self.id,
                    _admin_execution_models.ExecutionMetadata(
                        _admin_execution_models.ExecutionMetadata.ExecutionMode.MANUAL,
                        "sdk",  # TODO: get principle
                        0,  # TODO: Detect nesting
                    ),
                    notifications=notification_overrides,
                    disable_all=disable_all,
                    labels=label_overrides,
                    annotations=annotation_overrides,
                    auth_role=auth_role,
                ),
                literal_inputs,
            )
        except _user_exceptions.FlyteEntityAlreadyExistsException:
            exec_id = _identifier.WorkflowExecutionIdentifier(project, domain, name)
        execution = client.get_execution(exec_id)
        return _workflow_execution.SdkWorkflowExecution.promote_from_model(execution)



import os as _os
import uuid as _uuid
from distutils import dir_util as _dir_util
from shutil import copyfile as _copyfile

from flytekit.interfaces import random as _flyte_random
from flytekit.interfaces.data import common as _common_data


def _make_local_path(path):
    if not _os.path.exists(path):
        try:
            _os.makedirs(path)
        except OSError:  # Guard against race condition
            if not _os.path.isdir(path):
                raise


def strip_file_header(path: str) -> str:
    if path.startswith("file://"):
        return path.replace("file://", "", 1)
    return path


class LocalFileProxy(_common_data.DataProxy):
    def __init__(self, sandbox):
        """
        :param Text sandbox:
        """
        self._sandbox = sandbox

    @property
    def sandbox(self) -> str:
        return self._sandbox

    def exists(self, path):
        """
        :param Text path: the path of the file
        :rtype bool: whether the file exists or not
        """
        return _os.path.exists(strip_file_header(path))

    def download_directory(self, from_path, to_path):
        """
        :param Text from_path:
        :param Text to_path:
        """
        if from_path != to_path:
            _dir_util.copy_tree(strip_file_header(from_path), strip_file_header(to_path))

    def download(self, from_path, to_path):
        """
        :param Text from_path:
        :param Text to_path:
        """
        _copyfile(strip_file_header(from_path), strip_file_header(to_path))

    def upload(self, from_path, to_path):
        """
        :param Text from_path:
        :param Text to_path:
        """
        # Emulate s3's flat storage by automatically creating directory path
        _make_local_path(_os.path.dirname(strip_file_header(to_path)))
        # Write the object to a local file in the sandbox
        _copyfile(strip_file_header(from_path), strip_file_header(to_path))

    def upload_directory(self, from_path, to_path):
        """
        :param Text from_path:
        :param Text to_path:
        """
        self.download_directory(from_path, to_path)

    def get_random_path(self):
        """
        :rtype: Text
        """
        # Create a 128-bit random hash because the birthday attack principle shows that there is about a 50% chance of a
        # collision between objects when 2^(n/2) objects are created (where n is the number of bits in the hash).
        # Assuming Flyte eventually creates 1 trillion pieces of data (~2 ^ 40), the likelihood
        # of a collision is 10^-15 with 128-bit...or basically 0.
        return _os.path.join(self._sandbox, _uuid.UUID(int=_flyte_random.random.getrandbits(128)).hex)

    def get_random_directory(self):
        """
        :rtype: Text
        """
        random_dir = self.get_random_path() + "/"
        _make_local_path(random_dir)
        return random_dir



from flytekit.common.exceptions.base import FlyteException as _FlyteException
from flytekit.common.exceptions.base import FlyteRecoverableException as _Recoverable


class FlyteUserException(_FlyteException):
    _ERROR_CODE = "USER:Unknown"


class FlyteTypeException(FlyteUserException, TypeError):
    _ERROR_CODE = "USER:TypeError"

    @staticmethod
    def _is_a_container(value):
        return isinstance(value, list) or isinstance(value, tuple) or isinstance(value, set)

    @classmethod
    def _create_verbose_message(cls, received_type, expected_type, received_value=None, additional_msg=None):
        if received_value is not None:
            return "Type error!  Received: {} with value: {}, Expected{}: {}. {}".format(
                received_type,
                received_value,
                " one of" if FlyteTypeException._is_a_container(expected_type) else "",
                expected_type,
                additional_msg or "",
            )
        else:
            return "Type error!  Received: {}, Expected{}: {}. {}".format(
                received_type,
                " one of" if FlyteTypeException._is_a_container(expected_type) else "",
                expected_type,
                additional_msg or "",
            )

    def __init__(self, received_type, expected_type, additional_msg=None, received_value=None):
        super(FlyteTypeException, self).__init__(
            self._create_verbose_message(
                received_type,
                expected_type,
                received_value=received_value,
                additional_msg=additional_msg,
            )
        )


class FlyteValueException(FlyteUserException, ValueError):

    _ERROR_CODE = "USER:ValueError"

    @classmethod
    def _create_verbose_message(cls, received_value, error_message):
        return "Value error!  Received: {}. {}".format(received_value, error_message)

    def __init__(self, received_value, error_message):
        super(FlyteValueException, self).__init__(self._create_verbose_message(received_value, error_message))


class FlyteAssertion(FlyteUserException, AssertionError):
    _ERROR_CODE = "USER:AssertionError"


class FlyteValidationException(FlyteAssertion):
    _ERROR_CODE = "USER:ValidationError"


class FlyteEntityAlreadyExistsException(FlyteAssertion):
    _ERROR_CODE = "USER:EntityAlreadyExists"


class FlyteEntityNotExistException(FlyteAssertion):
    _ERROR_CODE = "USER:EntityNotExist"


class FlyteTimeout(FlyteAssertion):
    _ERROR_CODE = "USER:Timeout"


class FlyteRecoverableException(FlyteUserException, _Recoverable):
    _ERROR_CODE = "USER:Recoverable"


class FlyteAuthenticationException(FlyteAssertion):
    _ERROR_CODE = "USER:AuthenticationError"



import abc as _abc
import importlib as _importlib
import inspect as _inspect
import logging as _logging
from typing import Set

from flytekit.common import sdk_bases as _sdk_bases
from flytekit.common import utils as _utils
from flytekit.common.exceptions import system as _system_exceptions


class _InstanceTracker(_sdk_bases.ExtendedSdkType):
    """
    This is either genius or terrible.  Some of our tools iterate over modules and try to find Flyte entities
    (Tasks, Workflows, Launch Plans) and then register them.  However, if a task is imported via a command like:

        from package.module import some_task

    It is possible we will find a task reference twice, but then how do we know where it was defined?  Ideally, we would
    like to only register a task once and do so with the name where it is defined.  This metaclass allows us to do this
    by inspecting the call stack when __call__ is called on the metaclass (thus instantiating an object).
    """

    @staticmethod
    def _find_instance_module():
        frame = _inspect.currentframe()
        while frame:
            if frame.f_code.co_name == "<module>":
                return frame.f_globals["__name__"]
            frame = frame.f_back
        return None

    def __call__(cls, *args, **kwargs):
        o = super(_InstanceTracker, cls).__call__(*args, **kwargs)
        o._instantiated_in = _InstanceTracker._find_instance_module()
        return o


class FlyteEntity(object, metaclass=_sdk_bases.ExtendedSdkType):
    @property
    @_abc.abstractmethod
    def resource_type(self):
        """
        Integer from _identifier.ResourceType enum
        :rtype: int
        """
        pass

    @property
    @_abc.abstractmethod
    def entity_type_text(self):
        """
        TODO: Rename to resource type text
        :rtype: Text
        """
        pass


class TrackableEntity(FlyteEntity, metaclass=_InstanceTracker):
    def __init__(self, *args, **kwargs):
        self._platform_valid_name = None
        super(TrackableEntity, self).__init__(*args, **kwargs)

    @property
    def instantiated_in(self):
        """
        If found, we try to specify the module where the task was first instantiated.
        :rtype: Optional[Text]
        """
        return self._instantiated_in  # Set in metaclass

    @property
    def has_valid_name(self):
        """
        :rtype: bool
        """
        return self._platform_valid_name is not None and self._platform_valid_name != ""

    @property
    def platform_valid_name(self):
        """
        :rtype: Text
        """
        return self._platform_valid_name

    def assign_name(self, name):
        self._platform_valid_name = name

    def auto_assign_name(self):
        """
        This function is a bit of trickster Python code that goes hand in hand with the _InstanceTracker metaclass
        defined above. Thanks @matthewphsmith for this bit of ingenuity.

        For instance, if a user has code that looks like this:

            from some.other.module import wf
            my_launch_plan = wf.create_launch_plan()

            @dynamic_task
            def sample_task(wf_params):
                yield my_launch_plan()

        This code means that we should have a launch plan with a name ending in "my_launch_plan", since that is the
        name of the variable that the created launch plan gets assigned to. That is also the name that the launch plan
        would be registered with.

        However, when the create_launch_plan() function runs, the Python interpreter has no idea where the created
        object will be assigned to. It has no idea that the output of the create_launch_plan call is to be paired up
        with a variable named "my_launch_plan". This function basically does this after the fact. Leveraging the
        _instantiated_in field provided by the _InstanceTracker class above, this code will re-import the
        module (ie Python file) that the object is in. Since it's already loaded, it's just retrieved from memory.
        It then scans all objects in the module, and when an object match is found, it knows it's found the right
        variable name.

        Just to drive the point home, this function is mostly needed for Launch Plans. Assuming that user code has:

            @python_task
            def some_task()

        When Flytekit calls the module loader and loads the task, the name of the task is the name of the function
        itself.  It's known at time of creation. In contrast, when

            xyz = SomeWorkflow.create_launch_plan()

        is called, the name of the launch plan isn't known until after creation, it's not "SomeWorkflow", it's "xyz"
        """
        _logging.debug("Running name auto assign")
        m = _importlib.import_module(self.instantiated_in)

        for k in dir(m):
            try:
                if getattr(m, k) is self:
                    self._platform_valid_name = _utils.fqdn(m.__name__, k, entity_type=self.resource_type)
                    _logging.debug("Auto-assigning name to {}".format(self._platform_valid_name))
                    return
            except ValueError as err:
                # Empty pandas dataframes behave weirdly here such that calling `m.df` raises:
                # ValueError: The truth value of a {type(self).__name__} is ambiguous. Use a.empty, a.bool(), a.item(),
                #   a.any() or a.all()
                # Since dataframes aren't registrable entities to begin with we swallow any errors they raise and
                # continue looping through m.
                _logging.warning("Caught ValueError {} while attempting to auto-assign name".format(err))
                pass

        _logging.error("Could not auto-assign name")
        raise _system_exceptions.FlyteSystemException("Error looking for object while auto-assigning name.")


class RegisterableEntity(TrackableEntity):
    def __init__(self, *args, **kwargs):
        self._has_registered = False
        super(RegisterableEntity, self).__init__(*args, **kwargs)

    @_abc.abstractmethod
    def register(self, project, domain, name, version):
        """
        :param Text project: The project in which to register this task.
        :param Text domain: The domain in which to register this task.
        :param Text name: The name to give this task.
        :param Text version: The version in which to register this task.
        """
        pass

    @_abc.abstractmethod
    def serialize(self):
        """
        Registerable entities also are required to be serialized. This allows flytekit to separate serialization from
        the network call to Admin (mostly at least, if a Launch Plan is fetched for instance as part of another
        workflow, it will still hit Admin).
        """
        pass

    @property
    def has_registered(self) -> bool:
        return self._has_registered


class HasDependencies(object):
    """
    This interface is meant to describe Flyte entities that can have upstream dependencies. For instance, currently a
    launch plan depends on the underlying workflow, and a workflow is dependent on its tasks, and other launch plans,
    and subworkflows.
    """

    def __init__(self, *args, **kwargs):
        self._upstream_entities = set()
        super(HasDependencies, self).__init__(*args, **kwargs)

    @property
    def upstream_entities(self) -> Set[RegisterableEntity]:
        """
        Task, workflow, and launch plan that need to be registered in advance of this workflow.
        :rtype: set[RegisterableEntity]
        """
        return self._upstream_entities



import abc as _abc
import json as _json

import six as _six
from flyteidl.admin import common_pb2 as _common_pb2
from google.protobuf import json_format as _json_format
from google.protobuf import struct_pb2 as _struct


class FlyteABCMeta(_abc.ABCMeta):
    def __instancecheck__(cls, instance):
        if cls in type(instance).__mro__:
            return True
        return super(FlyteABCMeta, cls).__instancecheck__(instance)


class FlyteType(FlyteABCMeta):
    def __repr__(cls):
        return cls.short_class_string()

    def __str__(cls):
        return cls.verbose_class_string()

    def short_class_string(cls):
        """
        :rtype: Text
        """
        return super(FlyteType, cls).__repr__()

    def verbose_class_string(cls):
        """
        :rtype: Text
        """
        return cls.short_class_string()

    @_abc.abstractmethod
    def from_flyte_idl(cls, idl_object):
        pass


class FlyteIdlEntity(object, metaclass=FlyteType):
    def __eq__(self, other):
        return isinstance(other, FlyteIdlEntity) and other.to_flyte_idl() == self.to_flyte_idl()

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return self.short_string()

    def __str__(self):
        return self.verbose_string()

    def __hash__(self):
        return hash(self.to_flyte_idl().SerializeToString(deterministic=True))

    def short_string(self):
        """
        :rtype: Text
        """
        return _six.text_type(self.to_flyte_idl())

    def verbose_string(self):
        """
        :rtype: Text
        """
        return self.short_string()

    @property
    def is_empty(self):
        return len(self.to_flyte_idl().SerializeToString()) == 0

    @_abc.abstractmethod
    def to_flyte_idl(self):
        pass


class FlyteCustomIdlEntity(FlyteIdlEntity):
    @classmethod
    def from_flyte_idl(cls, idl_object):
        """

        :param _struct.Struct idl_object:
        :return: FlyteCustomIdlEntity
        """
        return cls.from_dict(idl_dict=_json_format.MessageToDict(idl_object))

    def to_flyte_idl(self):
        return _json_format.Parse(_json.dumps(self.to_dict()), _struct.Struct())

    @_abc.abstractmethod
    def from_dict(self, idl_dict):
        pass

    @_abc.abstractmethod
    def to_dict(self):
        """
        Converts self to a dictionary.
        :rtype: dict[Text, T]
        """
        pass


class NamedEntityIdentifier(FlyteIdlEntity):
    def __init__(self, project, domain, name=None):
        """
        :param Text project: The name of the project in which this entity lives.
        :param Text domain: The name of the domain within the project.
        :param Text name: [Optional] The name of the entity within the namespace of the project and domain.
        """
        self._project = project
        self._domain = domain
        self._name = name

    @property
    def project(self):
        """
        The name of the project in which this entity lives.
        :rtype: Text
        """
        return self._project

    @property
    def domain(self):
        """
        The name of the domain within the project.
        :rtype: Text
        """
        return self._domain

    @property
    def name(self):
        """
        The name of the entity within the namespace of the project and domain.
        :rtype: Text
        """
        return self._name

    def to_flyte_idl(self):
        """
        Stores object to a Flyte-IDL defined protobuf.
        :rtype: flyteidl.admin.common_pb2.NamedEntityIdentifier
        """

        # We use the kwarg constructor of the protobuf and setting name=None is equivalent to not setting it at all
        return _common_pb2.NamedEntityIdentifier(project=self.project, domain=self.domain, name=self.name)

    @classmethod
    def from_flyte_idl(cls, idl_object):
        """
        :param flyteidl.admin.common_pb2.NamedEntityIdentifier idl_object:
        :rtype: NamedEntityIdentifier
        """
        return cls(idl_object.project, idl_object.domain, idl_object.name)


class EmailNotification(FlyteIdlEntity):
    def __init__(self, recipients_email):
        """
        :param list[Text] recipients_email:
        """
        self._recipients_email = recipients_email

    @property
    def recipients_email(self):
        """
        :rtype: list[Text]
        """
        return self._recipients_email

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.common_pb2.EmailNotification
        """
        return _common_pb2.EmailNotification(recipients_email=self.recipients_email)

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.common_pb2.EmailNotification pb2_object:
        :rtype: EmailNotification
        """
        return cls(pb2_object.recipients_email)


class SlackNotification(FlyteIdlEntity):
    def __init__(self, recipients_email):
        """
        :param list[Text] recipients_email:
        """
        self._recipients_email = recipients_email

    @property
    def recipients_email(self):
        """
        :rtype: list[Text]
        """
        return self._recipients_email

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.common_pb2.SlackNotification
        """
        return _common_pb2.SlackNotification(recipients_email=self.recipients_email)

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.common_pb2.SlackNotification pb2_object:
        :rtype: EmailNotification
        """
        return cls(pb2_object.recipients_email)


class PagerDutyNotification(FlyteIdlEntity):
    def __init__(self, recipients_email):
        """
        :param list[Text] recipients_email:
        """
        self._recipients_email = recipients_email

    @property
    def recipients_email(self):
        """
        :rtype: list[Text]
        """
        return self._recipients_email

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.common_pb2.PagerDutyNotification
        """
        return _common_pb2.PagerDutyNotification(recipients_email=self.recipients_email)

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.common_pb2.PagerDutyNotification pb2_object:
        :rtype: EmailNotification
        """
        return cls(pb2_object.recipients_email)


class Notification(FlyteIdlEntity):
    def __init__(
        self,
        phases,
        email: EmailNotification = None,
        pager_duty: PagerDutyNotification = None,
        slack: SlackNotification = None,
    ):
        """
        Represents a structure for notifications based on execution status.

        :param list[int] phases: A list of phases to which users can associate the notifications.
        :param EmailNotification email: [Optional] Specify this for an email notification.
        :param PagerDutyNotification email: [Optional] Specify this for a PagerDuty notification.
        :param SlackNotification email: [Optional] Specify this for a Slack notification.
        """
        self._phases = phases
        self._email = email
        self._pager_duty = pager_duty
        self._slack = slack

    @property
    def phases(self):
        """
        A list of phases to which users can associate the notifications.
        :rtype: list[int]
        """
        return self._phases

    @property
    def email(self):
        """
        :rtype: EmailNotification
        """
        return self._email

    @property
    def pager_duty(self):
        """
        :rtype: PagerDutyNotification
        """
        return self._pager_duty

    @property
    def slack(self):
        """
        :rtype: SlackNotification
        """
        return self._slack

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.common_pb2.Notification
        """
        return _common_pb2.Notification(
            phases=self.phases,
            email=self.email.to_flyte_idl() if self.email else None,
            pager_duty=self.pager_duty.to_flyte_idl() if self.pager_duty else None,
            slack=self.slack.to_flyte_idl() if self.slack else None,
        )

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.admin.common_pb2.Notification p:
        :rtype: Notification
        """
        return cls(
            p.phases,
            email=EmailNotification.from_flyte_idl(p.email) if p.HasField("email") else None,
            pager_duty=PagerDutyNotification.from_flyte_idl(p.pager_duty) if p.HasField("pager_duty") else None,
            slack=SlackNotification.from_flyte_idl(p.slack) if p.HasField("slack") else None,
        )


class Labels(FlyteIdlEntity):
    def __init__(self, values):
        """
        Label values to be applied to a workflow execution resource.

        :param dict[Text, Text] values:
        """
        self._values = values

    @property
    def values(self):
        return self._values

    def to_flyte_idl(self):
        """
        :rtype: dict[Text, Text]
        """
        return _common_pb2.Labels(values={k: v for k, v in _six.iteritems(self.values)})

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.common_pb2.Labels pb2_object:
        :rtype: Labels
        """
        return cls({k: v for k, v in _six.iteritems(pb2_object.values)})


class Annotations(FlyteIdlEntity):
    def __init__(self, values):
        """
        Annotation values to be applied to a workflow execution resource.

        :param dict[Text, Text] values:
        """
        self._values = values

    @property
    def values(self):
        return self._values

    def to_flyte_idl(self):
        """
        :rtype: _common_pb2.Annotations
        """
        return _common_pb2.Annotations(values={k: v for k, v in _six.iteritems(self.values)})

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.common_pb2.Annotations pb2_object:
        :rtype: Annotations
        """
        return cls({k: v for k, v in _six.iteritems(pb2_object.values)})


class UrlBlob(FlyteIdlEntity):
    def __init__(self, url, bytes):
        """
        :param Text url:
        :param int bytes:
        """
        self._url = url
        self._bytes = bytes

    @property
    def url(self):
        """
        :rtype: Text
        """
        return self._url

    @property
    def bytes(self):
        """
        :rtype: int
        """
        return self._bytes

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.common_pb2.UrlBlob
        """
        return _common_pb2.UrlBlob(url=self.url, bytes=self.bytes)

    @classmethod
    def from_flyte_idl(cls, pb):
        """
        :param flyteidl.admin.common_pb2.UrlBlob pb:
        :rtype: UrlBlob
        """
        return cls(url=pb.url, bytes=pb.bytes)


class AuthRole(FlyteIdlEntity):
    def __init__(self, assumable_iam_role=None, kubernetes_service_account=None):
        """
        At most one of assumable_iam_role or kubernetes_service_account can be set.
        :param Text assumable_iam_role: IAM identity with set permissions policies.
        :param Text kubernetes_service_account: Provides an identity for workflow execution resources. Flyte deployment
            administrators are responsible for handling permissions as they relate to the service account.
        """
        if assumable_iam_role and kubernetes_service_account:
            raise ValueError("Only one of assumable_iam_role or kubernetes_service_account can be set")
        self._assumable_iam_role = assumable_iam_role
        self._kubernetes_service_account = kubernetes_service_account

    @property
    def assumable_iam_role(self):
        """
        The IAM role to execute the workflow with
        :rtype: Text
        """
        return self._assumable_iam_role

    @property
    def kubernetes_service_account(self):
        """
        The kubernetes service account to execute the workflow with
        :rtype: Text
        """
        return self._kubernetes_service_account

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.launch_plan_pb2.Auth
        """
        return _common_pb2.AuthRole(
            assumable_iam_role=self.assumable_iam_role if self.assumable_iam_role else None,
            kubernetes_service_account=self.kubernetes_service_account if self.kubernetes_service_account else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.launch_plan_pb2.Auth pb2_object:
        :rtype: Auth
        """
        return cls(
            assumable_iam_role=pb2_object.assumable_iam_role if pb2_object.HasField("assumable_iam_role") else None,
            kubernetes_service_account=pb2_object.kubernetes_service_account
            if pb2_object.HasField("kubernetes_service_account")
            else None,
        )


class RawOutputDataConfig(FlyteIdlEntity):
    def __init__(self, output_location_prefix):
        """
        :param Text output_location_prefix: Location of offloaded data for things like S3, etc.
        """
        self._output_location_prefix = output_location_prefix

    @property
    def output_location_prefix(self):
        return self._output_location_prefix

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.common_pb2.Auth
        """
        return _common_pb2.RawOutputDataConfig(output_location_prefix=self.output_location_prefix)

    @classmethod
    def from_flyte_idl(cls, pb2):
        return cls(output_location_prefix=pb2.output_location_prefix)



"""
Flytekit File Type (:mod:`flytekit.types.file`)
==========================================================
.. currentmodule:: flytekit.types.file

This list also contains a bunch of pre-formatted :py:class:`flytekit.types.file.FlyteFile` types.

.. autosummary::
   :toctree: generated/

   FlyteFile
   HDF5EncodedFile
   HTMLPage
   JoblibSerializedFile
   JPEGImageFile
   PDFFile
   PNGImageFile
   PythonPickledFile
   PythonNotebook
   SVGImageFile
"""

import typing

from .file import FlyteFile

# The following section provides some predefined aliases for commonly used FlyteFile formats.
# This makes their usage extremely simple for the users. Please keep the list sorted.


HDF5EncodedFile = FlyteFile[typing.TypeVar("hdf5")]
"""
    This can be used to denote that the returned file is of type hdf5 and can be received by other tasks that
    accept an hdf5 format. This is usually useful for serializing Tensorflow models
"""

HTMLPage = FlyteFile[typing.TypeVar("html")]
"""
    Can be used to receive or return an PNGImage. The underlying type is a FlyteFile, type. This is just a
    decoration and useful for attaching content type information with the file and automatically documenting code.
"""

JoblibSerializedFile = FlyteFile[typing.TypeVar("joblib")]
"""
    This File represents a file that was serialized using `joblib.dump` method can be loaded back using `joblib.load`
"""

JPEGImageFile = FlyteFile[typing.TypeVar("jpeg")]
"""
    Can be used to receive or return an JPEGImage. The underlying type is a FlyteFile, type. This is just a
    decoration and useful for attaching content type information with the file and automatically documenting code.
"""

PDFFile = FlyteFile[typing.TypeVar("pdf")]
"""
    Can be used to receive or return an PDFFile. The underlying type is a FlyteFile, type. This is just a
    decoration and useful for attaching content type information with the file and automatically documenting code.
"""

PNGImageFile = FlyteFile[typing.TypeVar("png")]
"""
    Can be used to receive or return an PNGImage. The underlying type is a FlyteFile, type. This is just a
    decoration and useful for attaching content type information with the file and automatically documenting code.
"""

PythonPickledFile = FlyteFile[typing.TypeVar("python-pickle")]
"""
    This type can be used when a serialized python pickled object is returned and shared between tasks. This only
    adds metadata to the file in Flyte, but does not really carry any object information
"""

PythonNotebook = FlyteFile[typing.TypeVar("ipynb")]
"""
    This type is used to identify a python notebook file
"""

SVGImageFile = FlyteFile[typing.TypeVar("svg")]
"""
    Can be used to receive or return an SVGImage. The underlying type is a FlyteFile, type. This is just a
    decoration and useful for attaching content type information with the file and automatically documenting code.
"""



from flyteidl.admin import common_pb2 as _common_pb2

from flytekit.models import common as _common


class Sort(_common.FlyteIdlEntity):
    class Direction(object):
        DESCENDING = _common_pb2.Sort.DESCENDING
        ASCENDING = _common_pb2.Sort.ASCENDING

    def __init__(self, key, direction):
        """
        :param Text key: field to sort on
        :param int direction: From flytekit.models.admin.common.Sort.Direction enum
        """
        self._key = key
        self._direction = direction

    @property
    def key(self):
        """
        :rtype: Text
        """
        return self._key

    @property
    def direction(self):
        """
        :rtype: int
        """
        return self._direction

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.common_pb2.Sort
        """
        return _common_pb2.Sort(key=self.key, direction=self.direction)

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.common_pb2.Sort pb2_object:
        :rtype: Sort
        """
        return cls(key=pb2_object.key, direction=pb2_object.direction)

    @classmethod
    def from_python_std(cls, text):
        """
        :param Text text:
        :rtype: Sort
        """
        text = text.strip()
        if text[-1] != ")":
            raise ValueError(
                "Could not parse string.  Must be in format 'asc(key)' or 'desc(key)'.  '{}' did not "
                "end with ')'.".format(text)
            )
        if text.startswith("asc("):
            direction = Sort.Direction.ASCENDING
            key = text[len("asc(") : -1].strip()
        elif text.startswith("desc("):
            direction = Sort.Direction.DESCENDING
            key = text[len("desc(") : -1].strip()
        else:
            raise ValueError(
                "Could not parse string.  Must be in format 'asc(key)' or 'desc(key)'.  '{}' did not "
                "start with 'asc(' or 'desc'.".format(text)
            )
        return cls(key=key, direction=direction)



import sys as _sys

import six as _six
from google.protobuf.json_format import MessageToDict as _MessageToDict

from flytekit import __version__
from flytekit.common import interface as _interface
from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.tasks import task as _base_tasks
from flytekit.common.types import helpers as _helpers
from flytekit.common.types import primitives as _primitives
from flytekit.configuration import internal as _internal_config
from flytekit.models import literals as _literal_models
from flytekit.models import task as _task_models

input_types_supported = {
    _primitives.Integer,
    _primitives.Boolean,
    _primitives.Float,
    _primitives.String,
    _primitives.Datetime,
    _primitives.Timedelta,
}


class SdkGenericSparkTask(_base_tasks.SdkTask):
    """
    This class includes the additional logic for building a task that executes as a Spark Job.

    """

    def __init__(
        self,
        task_type,
        discovery_version,
        retries,
        interruptible,
        task_inputs,
        deprecated,
        discoverable,
        timeout,
        spark_type,
        main_class,
        main_application_file,
        spark_conf,
        hadoop_conf,
        environment,
    ):
        """
        :param Text task_type: string describing the task type
        :param Text discovery_version: string describing the version for task discovery purposes
        :param int retries: Number of retries to attempt
        :param bool interruptible: Whether or not task is interruptible
        :param Text deprecated:
        :param bool discoverable:
        :param datetime.timedelta timeout:
        :param Text spark_type: Type of Spark Job: Scala/Java
        :param Text main_class: Main class to execute for Scala/Java jobs
        :param Text main_application_file: Main application file
        :param dict[Text,Text] spark_conf:
        :param dict[Text,Text] hadoop_conf:
        :param dict[Text,Text] environment: [optional] environment variables to set when executing this task.
        """

        spark_job = _task_models.SparkJob(
            spark_conf=spark_conf,
            hadoop_conf=hadoop_conf,
            spark_type=spark_type,
            application_file=main_application_file,
            main_class=main_class,
            executor_path=_sys.executable,
        ).to_flyte_idl()

        super(SdkGenericSparkTask, self).__init__(
            task_type,
            _task_models.TaskMetadata(
                discoverable,
                _task_models.RuntimeMetadata(
                    _task_models.RuntimeMetadata.RuntimeType.FLYTE_SDK,
                    __version__,
                    "spark",
                ),
                timeout,
                _literal_models.RetryStrategy(retries),
                interruptible,
                discovery_version,
                deprecated,
            ),
            _interface.TypedInterface({}, {}),
            _MessageToDict(spark_job),
        )

        # Add Inputs
        if task_inputs is not None:
            task_inputs(self)

        # Container after the Inputs have been updated.
        self._container = self._get_container_definition(environment=environment)

    def _validate_inputs(self, inputs):
        """
        :param dict[Text, flytekit.models.interface.Variable] inputs:  Input variables to validate
        :raises: flytekit.common.exceptions.user.FlyteValidationException
        """
        for k, v in _six.iteritems(inputs):
            sdk_type = _helpers.get_sdk_type_from_literal_type(v.type)
            if sdk_type not in input_types_supported:
                raise _user_exceptions.FlyteValidationException(
                    "Input Type '{}' not supported.  Only Primitives are supported for Scala/Java Spark.".format(
                        sdk_type
                    )
                )
        super(SdkGenericSparkTask, self)._validate_inputs(inputs)

    @_exception_scopes.system_entry_point
    def add_inputs(self, inputs):
        """
        Adds the inputs to this task.  This can be called multiple times, but it will fail if an input with a given
        name is added more than once, a name collides with an output, or if the name doesn't exist as an arg name in
        the wrapped function.
        :param dict[Text, flytekit.models.interface.Variable] inputs: names and variables
        """
        self._validate_inputs(inputs)
        self.interface.inputs.update(inputs)

    def _get_container_definition(
        self,
        environment=None,
    ):
        """
        :rtype: Container
        """

        args = []
        for k, v in _six.iteritems(self.interface.inputs):
            args.append("--{}".format(k))
            args.append("{{{{.Inputs.{}}}}}".format(k))

        return _task_models.Container(
            image=_internal_config.IMAGE.get(),
            command=[],
            args=args,
            resources=_task_models.Resources([], []),
            env=environment,
            config={},
        )



import logging as _logging
import os as _os
from datetime import datetime as _datetime

import six as _six
from flyteidl.plugins import qubole_pb2 as _qubole_pb2
from google.protobuf.json_format import ParseDict as _ParseDict
from six import moves as _six_moves

from flytekit.common import constants as _sdk_constants
from flytekit.common import utils as _common_utils
from flytekit.common.exceptions import system as _system_exception
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.types import helpers as _type_helpers
from flytekit.configuration import TemporaryConfiguration as _TemporaryConfiguration
from flytekit.engines import common as _common_engine
from flytekit.engines.unit.mock_stats import MockStats
from flytekit.interfaces.data import data_proxy as _data_proxy
from flytekit.models import array_job as _array_job
from flytekit.models import literals as _literals
from flytekit.models import qubole as _qubole_models
from flytekit.models.core.identifier import WorkflowExecutionIdentifier


class UnitTestEngineFactory(_common_engine.BaseExecutionEngineFactory):
    def get_task(self, sdk_task):
        """
        :param flytekit.common.tasks.task.SdkTask sdk_task:
        :rtype: UnitTestEngineTask
        """
        if sdk_task.type in {
            _sdk_constants.SdkTaskType.PYTHON_TASK,
            _sdk_constants.SdkTaskType.SPARK_TASK,
            _sdk_constants.SdkTaskType.SENSOR_TASK,
        }:
            return ReturnOutputsTask(sdk_task)
        elif sdk_task.type in {
            _sdk_constants.SdkTaskType.DYNAMIC_TASK,
        }:
            return DynamicTask(sdk_task)
        elif sdk_task.type in {
            _sdk_constants.SdkTaskType.BATCH_HIVE_TASK,
        }:
            return HiveTask(sdk_task)
        else:
            raise _user_exceptions.FlyteAssertion(
                "Unit tests are not currently supported for tasks of type: {}".format(sdk_task.type)
            )

    def get_workflow(self, _):
        raise _user_exceptions.FlyteAssertion("Unit testing of workflows is not currently supported")

    def get_launch_plan(self, _):
        raise _user_exceptions.FlyteAssertion("Unit testing of launch plans is not currently supported")

    def get_task_execution(self, _):
        raise _user_exceptions.FlyteAssertion("Unit testing does not return execution handles.")

    def get_node_execution(self, _):
        raise _user_exceptions.FlyteAssertion("Unit testing does not return execution handles.")

    def get_workflow_execution(self, _):
        raise _user_exceptions.FlyteAssertion("Unit testing does not return execution handles.")

    def fetch_workflow_execution(self, _):
        raise _user_exceptions.FlyteAssertion("Unit testing does not fetch execution handles.")

    def fetch_task(self, _):
        raise _user_exceptions.FlyteAssertion("Unit testing does not fetch real tasks.")

    def fetch_latest_task(self, named_task):
        raise _user_exceptions.FlyteAssertion("Unit testing does not fetch the real latest task.")

    def fetch_launch_plan(self, _):
        raise _user_exceptions.FlyteAssertion("Unit testing does not fetch real launch plans.")

    def fetch_workflow(self, _):
        raise _user_exceptions.FlyteAssertion("Unit testing does not fetch real workflows.")


class UnitTestEngineTask(_common_engine.BaseTaskExecutor):
    def execute(self, inputs, context=None):
        """
        Just execute the function and return the outputs as a user-readable dictionary.
        :param flytekit.models.literals.LiteralMap inputs:
        :param context:
        :rtype: dict[Text,flytekit.models.common.FlyteIdlEntity]
        """
        with _TemporaryConfiguration(
            _os.path.join(_os.path.dirname(__file__), "unit.config"),
            internal_overrides={"image": "unit_image"},
        ):
            with _common_utils.AutoDeletingTempDir("unit_test_dir") as working_directory:
                with _data_proxy.LocalWorkingDirectoryContext(working_directory):
                    return self._transform_for_user_output(self._execute_user_code(inputs))

    def _execute_user_code(self, inputs):
        """
        :param flytekit.models.literals.LiteralMap inputs:
        :rtype: dict[Text,flytekit.models.common.FlyteIdlEntity]
        """
        with _common_utils.AutoDeletingTempDir("user_dir") as user_working_directory:
            return self.sdk_task.execute(
                _common_engine.EngineContext(
                    execution_id=WorkflowExecutionIdentifier(project="unit_test", domain="unit_test", name="unit_test"),
                    execution_date=_datetime.utcnow(),
                    stats=MockStats(),
                    logging=_logging,  # TODO: A mock logging object that we can read later.
                    tmp_dir=user_working_directory,
                ),
                inputs,
            )

    def _transform_for_user_output(self, outputs):
        """
        Take whatever is returned from the task execution and convert to a reasonable output for the behavior of this
        task's unit test.
        :param dict[Text,flytekit.models.common.FlyteIdlEntity] outputs:
        :rtype: T
        """
        return outputs

    def register(self, identifier, version):
        raise _user_exceptions.FlyteAssertion("You cannot register unit test tasks.")

    def launch(
        self,
        project,
        domain,
        name=None,
        inputs=None,
        notification_overrides=None,
        label_overrides=None,
        annotation_overrides=None,
        auth_role=None,
    ):
        raise _user_exceptions.FlyteAssertion("You cannot launch unit test tasks.")


class ReturnOutputsTask(UnitTestEngineTask):
    def _transform_for_user_output(self, outputs):
        """
        Just return the outputs as a user-readable dictionary.
        :param dict[Text,flytekit.models.common.FlyteIdlEntity] outputs:
        :rtype: T
        """
        literal_map = outputs[_sdk_constants.OUTPUT_FILE_NAME]
        return {
            name: _type_helpers.get_sdk_value_from_literal(
                literal_map.literals[name],
                sdk_type=_type_helpers.get_sdk_type_from_literal_type(variable.type),
            ).to_python_std()
            for name, variable in _six.iteritems(self.sdk_task.interface.outputs)
        }


class DynamicTask(ReturnOutputsTask):
    def __init__(self, *args, **kwargs):
        self._has_workflow_node = False
        super(DynamicTask, self).__init__(*args, **kwargs)

    def _transform_for_user_output(self, outputs):
        if self.has_workflow_node:
            # If a workflow node has been detected, then we skip any transformation
            # This is to support the early termination behavior of the unit test engine when it comes to dynamic tasks
            # that produce launch plan or subworkflow nodes.
            # See the warning message in the code below for additional information
            return outputs
        return super(DynamicTask, self)._transform_for_user_output(outputs)

    def _execute_user_code(self, inputs):
        """
        :param flytekit.models.literals.LiteralMap inputs:
        :rtype: dict[Text,flytekit.models.common.FlyteIdlEntity]
        """
        results = super(DynamicTask, self)._execute_user_code(inputs)
        if _sdk_constants.FUTURES_FILE_NAME in results:
            futures = results[_sdk_constants.FUTURES_FILE_NAME]
            sub_task_outputs = {}
            tasks_map = {task.id: task for task in futures.tasks}

            for future_node in futures.nodes:
                if future_node.workflow_node is not None:
                    # TODO: implement proper unit testing for launchplan and subworkflow nodes somehow
                    _logging.warning(
                        "A workflow node has been detected in the output of the dynamic task. The "
                        "Flytekit unit test engine is incomplete for dynamic tasks that return launch "
                        "plans or subworkflows. The generated dynamic job spec will be returned but "
                        "they will not be run."
                    )
                    # For now, just return the output of the parent task
                    self._has_workflow_node = True
                    return results
                task = tasks_map[future_node.task_node.reference_id]
                if task.type == _sdk_constants.SdkTaskType.CONTAINER_ARRAY_TASK:
                    sub_task_output = DynamicTask.execute_array_task(future_node.id, task, results)
                elif task.type == _sdk_constants.SdkTaskType.SPARK_TASK:
                    # This is required because `_transform_for_user_output` function about is invoked which
                    # checks for outputs
                    self._has_workflow_node = True
                    return results
                elif task.type == _sdk_constants.SdkTaskType.HIVE_JOB:
                    # TODO: futures.outputs should have the Schema instances.
                    # After schema is implemented, fill out random data into the random locations
                    # then check output in test function
                    # Even though we recommend people use typed schemas, they might not always do so...
                    # in which case it'll be impossible to predict the actual schema, we should support a
                    # way for unit test authors to provide fake data regardless
                    sub_task_output = None
                else:
                    inputs_path = _os.path.join(future_node.id, _sdk_constants.INPUT_FILE_NAME)
                    if inputs_path not in results:
                        raise _system_exception.FlyteSystemAssertion(
                            "dynamic task hasn't generated expected inputs document [{}] found {}".format(
                                future_node.id, list(results.keys())
                            )
                        )
                    sub_task_output = UnitTestEngineFactory().get_task(task).execute(results[inputs_path])
                sub_task_outputs[future_node.id] = sub_task_output

            results[_sdk_constants.OUTPUT_FILE_NAME] = _literals.LiteralMap(
                literals={
                    binding.var: DynamicTask.fulfil_bindings(binding.binding, sub_task_outputs)
                    for binding in futures.outputs
                }
            )
        return results

    @property
    def has_workflow_node(self):
        """
        :rtype: bool
        """
        return self._has_workflow_node

    @staticmethod
    def execute_array_task(root_input_path, task, array_inputs):
        array_job = _array_job.ArrayJob.from_dict(task.custom)
        outputs = {}
        for job_index in _six_moves.range(0, array_job.size):
            inputs_path = _os.path.join(
                root_input_path,
                _six.text_type(job_index),
                _sdk_constants.INPUT_FILE_NAME,
            )
            if inputs_path not in array_inputs:
                raise _system_exception.FlyteSystemAssertion(
                    "dynamic task hasn't generated expected inputs document [{}].".format(inputs_path)
                )

            input_proto = array_inputs[inputs_path]
            # All outputs generated by the same array job will have the same key in sub_task_outputs,
            # they will, however, differ in the var names; they will be on the format [<job_index>].<var_name>
            # e.g. [1].out1
            for key, val in _six.iteritems(
                ReturnOutputsTask(
                    task.assign_type_and_return(_sdk_constants.SdkTaskType.PYTHON_TASK)  # TODO: This is weird
                ).execute(input_proto)
            ):
                outputs["[{}].{}".format(job_index, key)] = val
        return outputs

    @staticmethod
    def fulfil_bindings(binding_data, fulfilled_promises):
        """
        Substitutes promise values in binding_data with model Literal values built from python std values in
        fulfilled_promises

        :param _interface.BindingData binding_data:
        :param dict[Text,T] fulfilled_promises:
        :rtype:
        """
        if binding_data.scalar:
            return _literals.Literal(scalar=binding_data.scalar)
        elif binding_data.collection:
            return _literals.Literal(
                collection=_literals.LiteralCollection(
                    [
                        DynamicTask.fulfil_bindings(sub_binding_data, fulfilled_promises)
                        for sub_binding_data in binding_data.collection.bindings
                    ]
                )
            )
        elif binding_data.promise:
            if binding_data.promise.node_id not in fulfilled_promises:
                raise _system_exception.FlyteSystemAssertion(
                    "Expecting output of node [{}] but that hasn't been produced.".format(binding_data.promise.node_id)
                )
            node_output = fulfilled_promises[binding_data.promise.node_id]
            if binding_data.promise.var not in node_output:
                raise _system_exception.FlyteSystemAssertion(
                    "Expecting output [{}] of node [{}] but that hasn't been produced.".format(
                        binding_data.promise.var, binding_data.promise.node_id
                    )
                )

            return binding_data.promise.sdk_type.from_python_std(node_output[binding_data.promise.var])
        elif binding_data.map:
            return _literals.Literal(
                map=_literals.LiteralMap(
                    {
                        k: DynamicTask.fulfil_bindings(sub_binding_data, fulfilled_promises)
                        for k, sub_binding_data in _six.iteritems(binding_data.map.bindings)
                    }
                )
            )


class HiveTask(DynamicTask):
    def _transform_for_user_output(self, outputs):
        """
        Just execute the function and return the list of Hive queries returned.
        :param dict[Text,flytekit.models.common.FlyteIdlEntity] outputs:
        :rtype: list[Text]
        """
        futures = outputs.get(_sdk_constants.FUTURES_FILE_NAME)
        if futures:
            queries = []
            task_ids_to_defs = {
                t.id.name: _qubole_models.QuboleHiveJob.from_flyte_idl(
                    _ParseDict(t.custom, _qubole_pb2.QuboleHiveJob())
                )
                for t in futures.tasks
            }
            for node in futures.nodes:
                queries.append(task_ids_to_defs[node.task_node.reference_id.name].query.query)
            return queries
        else:
            return []



from dataclasses import dataclass
from enum import Enum
from typing import List, Optional

from flyteidl.core import security_pb2 as _sec

from flytekit.models import common as _common


@dataclass
class Secret(_common.FlyteIdlEntity):
    """
    Args:
        group is the Name of the secret. For example in kubernetes secrets is the name of the secret
        key is optional and can be an individual secret identifier within the secret For k8s this is required
        version is the version of the secret. This is an optional field
        mount_requirement provides a hint to the system as to how the secret should be injected
    """

    class MountType(Enum):
        ANY = _sec.Secret.MountType.ANY
        """
        Use this if the secret can be injected as either an environment variable / file and this should be left for the
        platform to decide. This is the most flexible option
        """
        ENV_VAR = _sec.Secret.MountType.ENV_VAR
        """
        Use this if the secret can be injected as an environment variable. Usually works for symmetric keys, passwords etc
        """
        FILE = _sec.Secret.MountType.FILE
        """
        Use this for Secrets that cannot be injected into env-var or need to be available as a file
        Caution: May not be supported in all environments
        """

    group: str
    key: str
    group_version: Optional[str] = None
    mount_requirement: MountType = MountType.ANY

    def __post_init__(self):
        if self.group is None:
            raise ValueError("Group is a required parameter")
        if self.key is None:
            raise ValueError("Key is also a required parameter")

    def to_flyte_idl(self) -> _sec.Secret:
        return _sec.Secret(
            group=self.group,
            group_version=self.group_version,
            key=self.key,
            mount_requirement=self.mount_requirement.value,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object: _sec.Secret) -> "Secret":
        return cls(
            group=pb2_object.group,
            group_version=pb2_object.group_version,
            key=pb2_object.key,
            mount_requirement=Secret.MountType(pb2_object.mount_requirement),
        )


@dataclass
class OAuth2Client(_common.FlyteIdlEntity):
    client_id: str
    client_secret: str

    def to_flyte_idl(self) -> _sec.OAuth2Client:
        return _sec.OAuth2Client(
            client_id=self.client_id,
            client_secret=self.client_secret,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object: _sec.OAuth2Client) -> "OAuth2Client":
        return cls(
            client_id=pb2_object.client_id,
            client_secret=pb2_object.client_secret,
        )


@dataclass
class Identity(_common.FlyteIdlEntity):
    iam_role: Optional[str] = None
    k8s_service_account: Optional[str] = None
    oauth2_client: Optional[OAuth2Client] = None

    def to_flyte_idl(self) -> _sec.Identity:
        return _sec.Identity(
            iam_role=self.iam_role if self.iam_role else None,
            k8s_service_account=self.k8s_service_account if self.k8s_service_account else None,
            oauth2_client=self.oauth2_client.to_flyte_idl() if self.oauth2_client else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object: _sec.Identity) -> "Identity":
        return cls(
            iam_role=pb2_object.iam_role,
            k8s_service_account=pb2_object.k8s_service_account,
            oauth2_client=OAuth2Client.from_flyte_idl(pb2_object.oauth2_client)
            if pb2_object.HasField("oauth2_client")
            else None,
        )


@dataclass
class OAuth2TokenRequest(_common.FlyteIdlEntity):
    class Type(Enum):
        CLIENT_CREDENTIALS = _sec.OAuth2TokenRequest.Type.CLIENT_CREDENTIALS

    name: str
    client: OAuth2Client
    idp_discovery_endpoint: Optional[str] = None
    token_endpoint: Optional[str] = None
    type_: Type = Type.CLIENT_CREDENTIALS

    def to_flyte_idl(self) -> _sec.OAuth2TokenRequest:
        return _sec.OAuth2TokenRequest(
            name=self.name,
            type=self.type_,
            token_endpoint=self.token_endpoint,
            idp_discovery_endpoint=self.idp_discovery_endpoint,
            client=self.client.to_flyte_idl() if self.client else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object: _sec.OAuth2TokenRequest) -> "OAuth2TokenRequest":
        return cls(
            name=pb2_object.name,
            idp_discovery_endpoint=pb2_object.idp_discovery_endpoint,
            token_endpoint=pb2_object.token_endpoint,
            type_=pb2_object.type,
            client=OAuth2Client.from_flyte_idl(pb2_object.client) if pb2_object.HasField("client") else None,
        )


@dataclass
class SecurityContext(_common.FlyteIdlEntity):
    run_as: Optional[Identity] = None
    secrets: Optional[List[Secret]] = None
    tokens: Optional[List[OAuth2TokenRequest]] = None

    def __post_init__(self):
        if self.secrets and not isinstance(self.secrets, list):
            self.secrets = [self.secrets]
        if self.tokens and not isinstance(self.tokens, list):
            self.tokens = [self.tokens]

    def to_flyte_idl(self) -> _sec.SecurityContext:
        if self.run_as is None and self.secrets is None and self.tokens is None:
            return None
        return _sec.SecurityContext(
            run_as=self.run_as.to_flyte_idl() if self.run_as else None,
            secrets=[s.to_flyte_idl() for s in self.secrets] if self.secrets else None,
            tokens=[t.to_flyte_idl() for t in self.tokens] if self.tokens else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object: _sec.SecurityContext) -> "SecurityContext":
        return cls(
            run_as=Identity.from_flyte_idl(pb2_object.run_as) if pb2_object.HasField("run_as") else None,
            secrets=[Secret.from_flyte_idl(s) for s in pb2_object.secrets] if pb2_object.secrets else None,
            tokens=[OAuth2TokenRequest.from_flyte_idl(t) for t in pb2_object.tokens] if pb2_object.tokens else None,
        )



import datetime as _datetime
from typing import List

from flytekit.common import constants as _constants
from flytekit.common import interface as _interface
from flytekit.common import nodes as _nodes
from flytekit.common import sdk_bases as _sdk_bases
from flytekit.common.core import identifier as _identifier
from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.exceptions import system as _system_exceptions
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.launch_plan import SdkLaunchPlan
from flytekit.common.mixins import hash as _hash_mixin
from flytekit.common.mixins import registerable as _registerable
from flytekit.configuration import auth as _auth_config
from flytekit.configuration import internal as _internal_config
from flytekit.engines.flyte import engine as _flyte_engine
from flytekit.models import common as _common_models
from flytekit.models import interface as _interface_models
from flytekit.models import launch_plan as _launch_plan_models
from flytekit.models import literals as _literal_models
from flytekit.models import schedule as _schedule_models
from flytekit.models.admin import workflow as _admin_workflow_model
from flytekit.models.core import identifier as _identifier_model
from flytekit.models.core import workflow as _workflow_models


class SdkWorkflow(
    _hash_mixin.HashOnReferenceMixin,
    _registerable.HasDependencies,
    _registerable.RegisterableEntity,
    _workflow_models.WorkflowTemplate,
    metaclass=_sdk_bases.ExtendedSdkType,
):
    """
    Previously this class represented both local and control plane constructs. As of this writing, we are making this
    class only a control plane class. Workflow constructs that rely on local code being present have been moved to
    the new PythonWorkflow class.
    """

    def __init__(
        self,
        nodes,
        interface,
        output_bindings,
        id,
        metadata,
        metadata_defaults,
    ):
        """
        :param list[flytekit.common.nodes.SdkNode] nodes:
        :param flytekit.models.interface.TypedInterface interface: Defines a strongly typed interface for the
            Workflow (inputs, outputs).  This can include some optional parameters.
        :param list[flytekit.models.literals.Binding] output_bindings: A list of output bindings that specify how to construct
            workflow outputs. Bindings can pull node outputs or specify literals. All workflow outputs specified in
            the interface field must be bound
            in order for the workflow to be validated. A workflow has an implicit dependency on all of its nodes
            to execute successfully in order to bind final outputs.
        :param flytekit.models.core.identifier.Identifier id: This is an autogenerated id by the system. The id is
            globally unique across Flyte.
        :param WorkflowMetadata metadata: This contains information on how to run the workflow.
        :param flytekit.models.core.workflow.WorkflowMetadataDefaults metadata_defaults: Defaults to be passed
            to nodes contained within workflow.
        """
        for n in nodes:
            for upstream in n.upstream_nodes:
                if upstream.id is None:
                    raise _user_exceptions.FlyteAssertion(
                        "Some nodes contained in the workflow were not found in the workflow description.  Please "
                        "ensure all nodes are either assigned to attributes within the class or an element in a "
                        "list, dict, or tuple which is stored as an attribute in the class."
                    )

        super(SdkWorkflow, self).__init__(
            id=id,
            metadata=metadata,
            metadata_defaults=metadata_defaults,
            interface=interface,
            nodes=nodes,
            outputs=output_bindings,
        )
        self._sdk_nodes = nodes
        self._has_registered = False

    @property
    def upstream_entities(self):
        return set(n.executable_sdk_object for n in self._sdk_nodes)

    @property
    def interface(self):
        """
        :rtype: flytekit.common.interface.TypedInterface
        """
        return super(SdkWorkflow, self).interface

    @property
    def entity_type_text(self):
        """
        :rtype: Text
        """
        return "Workflow"

    @property
    def resource_type(self):
        """
        Integer from _identifier.ResourceType enum
        :rtype: int
        """
        return _identifier_model.ResourceType.WORKFLOW

    def get_sub_workflows(self):
        """
        Recursive call that returns all subworkflows in the current workflow

        :rtype: list[SdkWorkflow]
        """
        result = []
        for node in self.nodes:
            if node.workflow_node is not None and node.workflow_node.sub_workflow_ref is not None:
                if node.executable_sdk_object is not None and node.executable_sdk_object.entity_type_text == "Workflow":
                    result.append(node.executable_sdk_object)
                    result.extend(node.executable_sdk_object.get_sub_workflows())
                else:
                    raise _system_exceptions.FlyteSystemException(
                        "workflow node with subworkflow found but bad executable "
                        "object {}".format(node.executable_sdk_object)
                    )

            # get subworkflows in conditional branches
            if node.branch_node is not None:
                if_else: _workflow_models.IfElseBlock = node.branch_node.if_else
                leaf_nodes: List[_nodes.SdkNode] = filter(
                    None,
                    [
                        if_else.case.then_node,
                        *([] if if_else.other is None else [x.then_node for x in if_else.other]),
                        if_else.else_node,
                    ],
                )
                for leaf_node in leaf_nodes:
                    exec_sdk_obj = leaf_node.executable_sdk_object
                    if exec_sdk_obj is not None and exec_sdk_obj.entity_type_text == "Workflow":
                        result.append(exec_sdk_obj)
                        result.extend(exec_sdk_obj.get_sub_workflows())

        return result

    @classmethod
    @_exception_scopes.system_entry_point
    def fetch(cls, project, domain, name, version=None):
        """
        This function uses the engine loader to call create a hydrated task from Admin.
        :param Text project:
        :param Text domain:
        :param Text name:
        :param Text version:
        :rtype: SdkWorkflow
        """
        version = version or _internal_config.VERSION.get()
        workflow_id = _identifier.Identifier(_identifier_model.ResourceType.WORKFLOW, project, domain, name, version)
        admin_workflow = _flyte_engine.get_client().get_workflow(workflow_id)
        cwc = admin_workflow.closure.compiled_workflow
        primary_template = cwc.primary.template
        sub_workflow_map = {sw.template.id: sw.template for sw in cwc.sub_workflows}
        task_map = {t.template.id: t.template for t in cwc.tasks}
        sdk_workflow = cls.promote_from_model(primary_template, sub_workflow_map, task_map)
        sdk_workflow._id = workflow_id
        sdk_workflow._has_registered = True
        return sdk_workflow

    @classmethod
    def get_non_system_nodes(cls, nodes):
        """
        :param list[flytekit.models.core.workflow.Node] nodes:
        :rtype: list[flytekit.models.core.workflow.Node]
        """
        return [n for n in nodes if n.id not in {_constants.START_NODE_ID, _constants.END_NODE_ID}]

    @classmethod
    def promote_from_model(cls, base_model, sub_workflows=None, tasks=None):
        """
        :param flytekit.models.core.workflow.WorkflowTemplate base_model:
        :param dict[flytekit.models.core.identifier.Identifier, flytekit.models.core.workflow.WorkflowTemplate]
            sub_workflows: Provide a list of WorkflowTemplate
            models (should be returned from Admin as part of the admin CompiledWorkflowClosure. Relevant sub-workflows
            should always be provided.
        :param dict[flytekit.models.core.identifier.Identifier, flytekit.models.task.TaskTemplate] tasks: Same as above
            but for tasks. If tasks are not provided relevant TaskTemplates will be fetched from Admin
        :rtype: SdkWorkflow
        """
        base_model_non_system_nodes = cls.get_non_system_nodes(base_model.nodes)
        sub_workflows = sub_workflows or {}
        tasks = tasks or {}
        node_map = {
            n.id: _nodes.SdkNode.promote_from_model(n, sub_workflows, tasks) for n in base_model_non_system_nodes
        }

        # Set upstream nodes for each node
        for n in base_model_non_system_nodes:
            current = node_map[n.id]
            for upstream_id in current.upstream_node_ids:
                upstream_node = node_map[upstream_id]
                current << upstream_node

        # No inputs/outputs specified, see the constructor for more information on the overrides.
        return cls(
            nodes=list(node_map.values()),
            id=_identifier.Identifier.promote_from_model(base_model.id),
            metadata=base_model.metadata,
            metadata_defaults=base_model.metadata_defaults,
            interface=_interface.TypedInterface.promote_from_model(base_model.interface),
            output_bindings=base_model.outputs,
        )

    @_exception_scopes.system_entry_point
    def register(self, project, domain, name, version):
        """
        :param Text project:
        :param Text domain:
        :param Text name:
        :param Text version:
        """
        self.validate()
        id_to_register = _identifier.Identifier(_identifier_model.ResourceType.WORKFLOW, project, domain, name, version)
        old_id = self.id
        self._id = id_to_register
        try:
            client = _flyte_engine.get_client()
            sub_workflows = self.get_sub_workflows()
            client.create_workflow(
                id_to_register,
                _admin_workflow_model.WorkflowSpec(
                    self,
                    sub_workflows,
                ),
            )
            self._id = id_to_register
            self._has_registered = True
            return str(id_to_register)
        except _user_exceptions.FlyteEntityAlreadyExistsException:
            pass
        except Exception:
            self._id = old_id
            raise

    @_exception_scopes.system_entry_point
    def serialize(self):
        """
        Serializing a workflow should produce an object similar to what the registration step produces, in preparation
        for actual registration to Admin.

        :rtype: flyteidl.admin.workflow_pb2.WorkflowSpec
        """
        sub_workflows = self.get_sub_workflows()
        return _admin_workflow_model.WorkflowSpec(
            self,
            sub_workflows,
        ).to_flyte_idl()

    @_exception_scopes.system_entry_point
    def validate(self):
        pass

    @_exception_scopes.system_entry_point
    def create_launch_plan(self, *args, **kwargs):
        # TODO: Correct after implementing new launch plan
        assumable_iam_role = _auth_config.ASSUMABLE_IAM_ROLE.get()
        kubernetes_service_account = _auth_config.KUBERNETES_SERVICE_ACCOUNT.get()

        if not (assumable_iam_role or kubernetes_service_account):
            raise _user_exceptions.FlyteValidationException("No assumable role or service account found")
        auth_role = _common_models.AuthRole(
            assumable_iam_role=assumable_iam_role,
            kubernetes_service_account=kubernetes_service_account,
        )

        return SdkLaunchPlan(
            workflow_id=self.id,
            entity_metadata=_launch_plan_models.LaunchPlanMetadata(
                schedule=_schedule_models.Schedule(""),
                notifications=[],
            ),
            default_inputs=_interface_models.ParameterMap({}),
            fixed_inputs=_literal_models.LiteralMap(literals={}),
            labels=_common_models.Labels({}),
            annotations=_common_models.Annotations({}),
            auth_role=auth_role,
            raw_output_data_config=_common_models.RawOutputDataConfig(""),
        )

    @_exception_scopes.system_entry_point
    def __call__(self, *args, **input_map):
        if len(args) > 0:
            raise _user_exceptions.FlyteAssertion(
                "When adding a workflow as a node in a workflow, all inputs must be specified with kwargs only.  We "
                "detected {} positional args.".format(len(args))
            )
        bindings, upstream_nodes = self.interface.create_bindings_for_inputs(input_map)

        node = _nodes.SdkNode(
            id=None,
            metadata=_workflow_models.NodeMetadata(
                "placeholder", _datetime.timedelta(), _literal_models.RetryStrategy(0)
            ),
            upstream_nodes=upstream_nodes,
            bindings=sorted(bindings, key=lambda b: b.var),
            sdk_workflow=self,
        )
        return node



from collections import OrderedDict

import pandas
import pytest
from flytekitplugins.hive.task import HiveConfig, HiveSelectTask, HiveTask

from flytekit import kwtypes, workflow
from flytekit.extend import Image, ImageConfig, SerializationSettings, get_serializable
from flytekit.testing import task_mock
from flytekit.types.schema import FlyteSchema


def test_serialization():
    hive_task = HiveTask(
        name="flytekit.demo.hive_task.hivequery1",
        inputs=kwtypes(my_schema=FlyteSchema, ds=str),
        config=HiveConfig(cluster_label="flyte"),
        query_template="""
            set engine=tez;
            insert overwrite directory '{{ .rawOutputDataPrefix }}' stored as parquet  -- will be unique per retry
            select *
            from blah
            where ds = '{{ .Inputs.ds }}' and uri = '{{ .inputs.my_schema }}'
        """,
        # the schema literal's backend uri will be equal to the value of .raw_output_data
        output_schema_type=FlyteSchema,
    )

    @workflow
    def my_wf(in_schema: FlyteSchema, ds: str) -> FlyteSchema:
        return hive_task(my_schema=in_schema, ds=ds)

    default_img = Image(name="default", fqn="test", tag="tag")
    serialization_settings = SerializationSettings(
        project="proj",
        domain="dom",
        version="123",
        image_config=ImageConfig(default_image=default_img, images=[default_img]),
        env={},
    )
    sdk_task = get_serializable(OrderedDict(), serialization_settings, hive_task)
    assert "{{ .rawOutputDataPrefix" in sdk_task.custom["query"]["query"]
    assert "insert overwrite directory" in sdk_task.custom["query"]["query"]
    assert len(sdk_task.interface.inputs) == 2
    assert len(sdk_task.interface.outputs) == 1

    sdk_wf = get_serializable(OrderedDict(), serialization_settings, my_wf)
    assert sdk_wf.interface.outputs["o0"].type.schema is not None
    assert sdk_wf.outputs[0].var == "o0"
    assert sdk_wf.outputs[0].binding.promise.node_id == "n0"
    assert sdk_wf.outputs[0].binding.promise.var == "results"


def test_local_exec():
    hive_task = HiveTask(
        name="flytekit.demo.hive_task.hivequery1",
        inputs={},
        config=HiveConfig(cluster_label="flyte"),
        query_template="""
            set engine=tez;
            insert overwrite directory '{{ .raw_output_data }}' stored as parquet  -- will be unique per retry
            select *
            from blah
            where ds = '{{ .Inputs.ds }}' and uri = '{{ .inputs.my_schema }}'
        """,
        output_schema_type=FlyteSchema,
    )

    assert len(hive_task.interface.inputs) == 0
    assert len(hive_task.interface.outputs) == 1

    # will not run locally
    with pytest.raises(Exception):
        hive_task()

    my_demo_output = pandas.DataFrame(data={"x": [1, 2], "y": ["3", "4"]})

    @workflow
    def my_wf() -> FlyteSchema:
        return hive_task()

    with task_mock(hive_task) as mock:
        mock.return_value = my_demo_output
        x = my_wf()
        df = x.open().all()
        y = df == my_demo_output
        assert y.all().all()


def test_query_no_inputs_or_outputs():
    hive_task = HiveTask(
        name="flytekit.demo.hive_task.hivequery1",
        inputs={},
        task_config=HiveConfig(cluster_label="flyte"),
        query_template="""
            insert into extant_table (1, 'two')
        """,
        output_schema_type=None,
    )

    @workflow
    def my_wf():
        hive_task()

    default_img = Image(name="default", fqn="test", tag="tag")
    serialization_settings = SerializationSettings(
        project="proj",
        domain="dom",
        version="123",
        image_config=ImageConfig(default_image=default_img, images=[default_img]),
        env={},
    )
    sdk_task = get_serializable(OrderedDict(), serialization_settings, hive_task)
    assert len(sdk_task.interface.inputs) == 0
    assert len(sdk_task.interface.outputs) == 0

    get_serializable(OrderedDict(), serialization_settings, my_wf)


def test_hive_select():
    hive_select = HiveSelectTask(
        name="flytekit.demo.hive_task.hivequery1",
        inputs={},
        task_config=HiveConfig(cluster_label="flyte"),
        select_query="select 1, 2, 3",
        output_schema_type=FlyteSchema,
    )

    sql = hive_select.get_query()
    assert "{{ .PerRetryUniqueKey }}_tmp" in sql



import os as _os
from typing import Any, Dict, List

from flyteidl.core import literals_pb2 as _literals_pb2

from flytekit.clients.helpers import iterate_node_executions as _iterate_node_executions
from flytekit.common import utils as _common_utils
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.mixins import artifact as _artifact
from flytekit.control_plane import identifier as _core_identifier
from flytekit.control_plane import nodes as _nodes
from flytekit.engines.flyte import engine as _flyte_engine
from flytekit.interfaces.data import data_proxy as _data_proxy
from flytekit.models import execution as _execution_models
from flytekit.models import filters as _filter_models
from flytekit.models import literals as _literal_models
from flytekit.models.core import execution as _core_execution_models


class FlyteWorkflowExecution(_execution_models.Execution, _artifact.ExecutionArtifact):
    def __init__(self, *args, **kwargs):
        super(FlyteWorkflowExecution, self).__init__(*args, **kwargs)
        self._node_executions = None
        self._inputs = None
        self._outputs = None

    @property
    def node_executions(self) -> Dict[str, _nodes.FlyteNodeExecution]:
        return self._node_executions or {}

    @property
    def inputs(self) -> Dict[str, Any]:
        """
        Returns the inputs to the execution in the standard python format as dictated by the type engine.
        """
        if self._inputs is None:
            client = _flyte_engine.get_client()
            execution_data = client.get_execution_data(self.id)

            # Inputs are returned inline unless they are too big, in which case a url blob pointing to them is returned.
            input_map: LiteralMap = _literal_models.LiteralMap({})
            if bool(execution_data.full_inputs.literals):
                input_map = execution_data.full_inputs
            elif execution_data.inputs.bytes > 0:
                with _common_utils.AutoDeletingTempDir() as tmp_dir:
                    tmp_name = _os.path.join(tmp_dir.name, "inputs.pb")
                    _data_proxy.Data.get_data(execution_data.inputs.url, tmp_name)
                    input_map = _literal_models.LiteralMap.from_flyte_idl(
                        _common_utils.load_proto_from_file(_literals_pb2.Literalmap, tmp_name)
                    )
            # TODO: need to convert flyte literals to python types. For now just use literals
            # self._inputs = TypeEngine.literal_map_to_kwargs(ctx=FlyteContext.current_context(), lm=input_map)
            self._inputs = input_map
        return self._inputs

    @property
    def outputs(self) -> Dict[str, Any]:
        """
        Returns the outputs to the execution in the standard python format as dictated by the type engine.

        :raises: ``FlyteAssertion`` error if execution is in progress or execution ended in error.
        """
        if not self.is_complete:
            raise _user_exceptions.FlyteAssertion(
                "Please wait until the node execution has completed before requesting the outputs."
            )
        if self.error:
            raise _user_exceptions.FlyteAssertion("Outputs could not be found because the execution ended in failure.")

        if self._outputs is None:
            client = _flyte_engine.get_client()
            execution_data = client.get_execution_data(self.id)
            # Outputs are returned inline unless they are too big, in which case a url blob pointing to them is returned.
            output_map: LiteralMap = _literal_models.LiteralMap({})
            if bool(execution_data.full_outputs.literals):
                output_map = execution_data.full_outputs
            elif execution_data.outputs.bytes > 0:
                with _common_utils.AutoDeletingTempDir() as tmp_dir:
                    tmp_name = _os.path.join(tmp_dir.name, "outputs.pb")
                    _data_proxy.Data.get_data(execution_data.outputs.url, tmp_name)
                    output_map = _literal_models.LiteralMap.from_flyte_idl(
                        _common_utils.load_proto_from_file(_literals_pb2.LiteralMap, tmp_name)
                    )
            # TODO: need to convert flyte literals to python types. For now just use literals
            # self._outputs = TypeEngine.literal_map_to_kwargs(ctx=FlyteContext.current_context(), lm=output_map)
            self._outputs = output_map
        return self._outputs

    @property
    def error(self) -> _core_execution_models.ExecutionError:
        """
        If execution is in progress, raise an exception.  Otherwise, return None if no error was present upon
        reaching completion.
        """
        if not self.is_complete:
            raise _user_exceptions.FlyteAssertion(
                "Please wait until a workflow has completed before checking for an " "error."
            )
        return self.closure.error

    @property
    def is_complete(self) -> bool:
        """
        Whether or not the execution is complete.
        """
        return self.closure.phase in {
            _core_execution_models.WorkflowExecutionPhase.ABORTED,
            _core_execution_models.WorkflowExecutionPhase.FAILED,
            _core_execution_models.WorkflowExecutionPhase.SUCCEEDED,
            _core_execution_models.WorkflowExecutionPhase.TIMED_OUT,
        }

    @classmethod
    def promote_from_model(cls, base_model: _execution_models.Execution) -> "FlyteWorkflowExecution":
        return cls(
            closure=base_model.closure,
            id=_core_identifier.WorkflowExecutionIdentifier.promote_from_model(base_model.id),
            spec=base_model.spec,
        )

    @classmethod
    def fetch(cls, project: str, domain: str, name: str) -> "FlyteWorkflowExecution":
        return cls.promote_from_model(
            _flyte_engine.get_client().get_execution(
                _core_identifier.WorkflowExecutionIdentifier(project=project, domain=domain, name=name)
            )
        )

    def sync(self):
        """
        Syncs the state of the underlying execution artifact with the state observed by the platform.
        """
        if not self.is_complete or self._node_executions is None:
            self._sync_closure()
            self._node_executions = self.get_node_executions()

    def _sync_closure(self):
        if not self.is_complete:
            client = _flyte_engine.get_client()
            self._closure = client.get_execution(self.id).closure

    def get_node_executions(self, filters: List[_filter_models.Filter] = None) -> Dict[str, _nodes.FlyteNodeExecution]:
        client = _flyte_engine.get_client()
        return {
            node.id.node_id: _nodes.FlyteNodeExecution.promote_from_model(node)
            for node in _iterate_node_executions(client, self.id, filters=filters)
        }

    def terminate(self, cause: str):
        _flyte_engine.get_client().terminate_execution(self.id, cause)



from typing import Any, Dict, List, Tuple

from flytekit.control_plane import nodes as _nodes
from flytekit.models import interface as _interface_models
from flytekit.models import literals as _literal_models


class TypedInterface(_interface_models.TypedInterface):
    @classmethod
    def promote_from_model(cls, model):
        """
        :param flytekit.models.interface.TypedInterface model:
        :rtype: TypedInterface
        """
        return cls(model.inputs, model.outputs)

    def create_bindings_for_inputs(
        self, map_of_bindings: Dict[str, Any]
    ) -> Tuple[List[_literal_models.Binding], List[_nodes.FlyteNode]]:
        """
        :param: map_of_bindings: this can be scalar primitives, it can be node output references, lists, etc.
        :raises: flytekit.common.exceptions.user.FlyteAssertion
        """
        return [], []



import logging as _logging
import urllib.parse as _urlparse

from flytekit.clis.auth.auth import AuthorizationClient as _AuthorizationClient
from flytekit.clis.auth.discovery import DiscoveryClient as _DiscoveryClient
from flytekit.configuration.creds import CLIENT_ID as _CLIENT_ID
from flytekit.configuration.creds import REDIRECT_URI as _REDIRECT_URI
from flytekit.configuration.platform import HTTP_URL as _HTTP_URL
from flytekit.configuration.platform import INSECURE as _INSECURE
from flytekit.configuration.platform import URL as _URL

# Default, well known-URI string used for fetching JSON metadata. See https://tools.ietf.org/html/rfc8414#section-3.
discovery_endpoint_path = "./.well-known/oauth-authorization-server"


def _get_discovery_endpoint(http_config_val, platform_url_val, insecure_val):

    if http_config_val:
        scheme, netloc, path, _, _, _ = _urlparse.urlparse(http_config_val)
        if not scheme:
            scheme = "http" if insecure_val else "https"
    else:  # Use the main _URL config object effectively
        scheme = "http" if insecure_val else "https"
        netloc = platform_url_val
        path = ""

    computed_endpoint = _urlparse.urlunparse((scheme, netloc, path, None, None, None))
    # The urljoin function needs a trailing slash in order to append things correctly. Also, having an extra slash
    # at the end is okay, it just gets stripped out.
    computed_endpoint = _urlparse.urljoin(computed_endpoint + "/", discovery_endpoint_path)
    _logging.info("Using {} as discovery endpoint".format(computed_endpoint))
    return computed_endpoint


# Lazy initialized authorization client singleton
_authorization_client = None


def get_client(flyte_client_url):
    global _authorization_client
    if _authorization_client is not None and not _authorization_client.expired:
        return _authorization_client
    authorization_endpoints = get_authorization_endpoints(flyte_client_url)

    _authorization_client = _AuthorizationClient(
        redirect_uri=_REDIRECT_URI.get(),
        client_id=_CLIENT_ID.get(),
        auth_endpoint=authorization_endpoints.auth_endpoint,
        token_endpoint=authorization_endpoints.token_endpoint,
    )
    return _authorization_client


def get_authorization_endpoints(flyte_client_url):
    discovery_endpoint = _get_discovery_endpoint(_HTTP_URL.get(), flyte_client_url or _URL.get(), _INSECURE.get())
    discovery_client = _DiscoveryClient(discovery_url=discovery_endpoint)
    return discovery_client.get_authorization_endpoints()



from flytekit.common.types import blobs as _blobs
from flytekit.common.types import containers as _containers
from flytekit.common.types import helpers as _helpers
from flytekit.common.types import primitives as _primitives
from flytekit.common.types import proto as _proto
from flytekit.common.types import schema as _schema


class Types(object):
    Integer = _helpers.get_sdk_type_from_literal_type(_primitives.Integer.to_flyte_literal_type())
    """
    Use this to specify a simple integer type.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            1) If set, A Python int will be received, if set.
            2) Otherwise, a None value will be received.

        As output:
            1) User code may pass an int or long value.
            2) Output can also be nulled with a None value.

        From command-line:
            Specify an integer or integer string.

    .. code-block:: python

        @inputs(a=Types.Integer)
        @outputs(b=Types.Integer)
        @python_task
        def double(wf_params, a, b):
            b.set(a * 2)
    """

    Float = _helpers.get_sdk_type_from_literal_type(_primitives.Float.to_flyte_literal_type())
    """
    Use this to specify a simple floating point type.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            A Python float will be received, if set.  Otherwise, a None value will be received.

        As output:
            User code may pass a float value.  It can also be nulled with a None value.

        From command-line:
            Specify a float or floating-point string.

    .. code-block:: python

        @inputs(a=Types.Float)
        @outputs(b=Types.Float)
        @python_task
        def square(wf_params, a, b):
            b.set(a * a)
    """

    String = _helpers.get_sdk_type_from_literal_type(_primitives.String.to_flyte_literal_type())
    """
    Use this to specify a simple string type.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            A Python str (Python 2) or unicode (Python 3) will be received, if set.  Otherwise, a None value will be
            received.

        As output:
            User code may pass a str value (Python 2) or a unicode value (Python 3).  It can also be nulled with a None
            value.

        From command-line:
            Specify a string.

    .. code-block:: python

        @inputs(a=Types.String, b=Types.String)
        @outputs(c=Types.String)
        @python_task
        def concat(wf_params, a, b):
            c.set(a + b)
    """

    Boolean = _helpers.get_sdk_type_from_literal_type(_primitives.Boolean.to_flyte_literal_type())
    """
    Use this to specify a simple bool type.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            A Python bool will be received, if set.  Otherwise, a None value will be received.

        As output:
            User code may pass a bool value.  It can also be nulled with a None value.

        From command-line:
            Specify 0, 1, true, or false.

    .. code-block:: python

        @inputs(a=Types.Boolean)
        @outputs(b=Types.Boolean)
        @python_task
        def invert(wf_params, a, b):
            b.set(not a)
    """

    Datetime = _helpers.get_sdk_type_from_literal_type(_primitives.Datetime.to_flyte_literal_type())
    """
    Use this to specify a simple datetime type.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            A Python timezone-aware datetime.datetime will be received with a UTC time, if set.  Otherwise,
            a None value will be received.

        As output:
            User code may pass a timezone-aware datetime.datetime value.  It can also be nulled with a None value.

        From command-line:
            Specify a timezone-aware, parsable datestring.  i.e. 2019-01-01T00:00+00:00

    .. note::

        The engine requires that datetimes be timezone aware.  By default, Python datetime.datetime is not timezone
        aware.

    .. code-block:: python

        @inputs(a=Types.Datetime)
        @outputs(b=Types.Datetime)
        @python_task
        def tomorrow(wf_params, a, b):
            b.set(a + datetime.timedelta(days=1))
    """

    Timedelta = _helpers.get_sdk_type_from_literal_type(_primitives.Timedelta.to_flyte_literal_type())
    """
    Use this to specify a simple timedelta type.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            A Python datetime.timedelta will be received, if set.  Otherwise, a None value will be received.

        As output:
            User code may pass a datetime.timedelta value.  It can also be nulled with a None value.

        From command-line:
            Specify a parsable duration string.  i.e. 1h30m24s

    .. code-block:: python

        @inputs(a=Types.Timedelta)
        @outputs(b=Types.Timedelta)
        @python_task
        def hundred_times_longer(wf_params, a, b):
            b.set(a * 100)
    """

    Generic = _helpers.get_sdk_type_from_literal_type(_primitives.Generic.to_flyte_literal_type())
    """
    Use this to specify a simple JSON type. The Generic type offer a flexible (but loose) extension to flyte's typing
    system by allowing custom types/objects to be passed through. It's strongly recommended for producers & consumers of
    entities that produce or consume a Generic type to perform their own expectations checks on the integrity of the
    object.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            1) If set, a Python dict with JSON-ifiable primitives and nested lists or maps.
            2) Otherwise, a None value will be received.

        As output:
            1) User code may pass a Python dict with arbitrarily nested lists and dictionaries.  JSON-ifiable
               primitives may also be specified.
            2) Output can also be nulled with a None value.

        From command-line:
            Specify a JSON string.

    .. code-block:: python

        @inputs(a=Types.Generic)
        @outputs(b=Types.Generic)
        @python_task
        def operate(wf_params, a, b):
            if a['operation'] == 'add':
                a['value'] += a['operand']  # a['value'] is a number
            elif a['operation'] == 'merge':
                a['value'].update(a['some']['nested'][0]['field'])
            b.set(a)

        # For better readability, it's strongly advised to leverage python's type aliasing.
        MyTypeA = Types.Generic
        MyTypeB = Types.Generic

        # This makes it clearer that it received a certain type and produces a different one. Other tasks that consume
        # MyTypeB should do so in their input declaration.
        @inputs(a=MyTypeA)
        @outputs(b=MyTypeB)
        @python_task
        def operate(wf_params, a, b):
            if a['operation'] == 'add':
                a['value'] += a['operand']  # a['value'] is a number
            elif a['operation'] == 'merge':
                a['value'].update(a['some']['nested'][0]['field'])
            b.set(a)
    """

    Blob = _blobs.Blob
    """
    Use this to specify a Blob object which is essentially a managed file.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            1) If set, a :py:class:`flytekit.common.types.impl.blobs.Blob` object will be received.
            2) If not set, a None value.

        As output:
            1) A user may specify a path string.
            2) A user may construct a :py:class:`flytekit.common.types.impl.blobs.Blob` object and pass it as output.
            3) Output can be nulled with a None value.

        From command-line:
            Specify a path to the blob. This path must be accessible from the container when executing--either by
            being downloaded from an accessible remote location like s3 or as a local file.

    .. code-block:: python

        @inputs(a=Types.Blob)
        @outputs(b=Types.Blob)
        @python_task
        def copy(wf_params, a, b):
            with a as reader:
                txt = reader.read()

            out = Types.Blob()  # Create at a random location specified in flytekit configuration
            with out as writer:
                writer.write(txt)
            b.set(out)
    """

    CSV = _blobs.CSV
    """
    Use this to specify a CSV blob object which is essentially a managed file in the CSV format.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            1) If set, a :py:class:`flytekit.common.types.impl.blobs.CSV` object will be received.
            2) If not set, a None value.

        As output:
            1) A user may specify a path string.
            2) A user may construct a :py:class:`flytekit.common.types.impl.blobs.CSV` object and pass it as output.
            3) Output can be nulled with a None value.

        From command-line:
            Specify a path to the CSV. This path must be accessible from the container when executing--either by
            being downloaded from an accessible remote location like s3 or as a local file.

    .. code-block:: python

        @inputs(a=Types.CSV)
        @outputs(b=Types.CSV)
        @python_task
        def copy(wf_params, a, b):
            with a as reader:
                txt = reader.read()

            out = Types.CSV()  # Create at a random location specified in flytekit configuration
            with out as writer:
                writer.write(txt)
            b.set(out)
    """

    MultiPartBlob = _blobs.MultiPartBlob
    """
    Use this to specify a multi-part blob object which is essentially a chunked file in a non-recursive directory.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            1) If set, a :py:class:`flytekit.common.types.impl.blobs.MultiPartBlob` object will be received.
            2) If not set, a None value.

        As output:
            1) A user may specify a path string.
            2) A user may construct a :py:class:`flytekit.common.types.impl.blobs.MultiPartBlob` object and pass it as
               output.
            3) Output can be nulled with a None value.

        From command-line:
            Specify a path to the multi-part blob. This path must be accessible from the container when
            executing--either by being downloaded from an accessible remote location like s3 or as a local file.

    .. code-block:: python

        @inputs(a=Types.MultiPartBlob)
        @outputs(b=Types.MultiPartBlob)
        @python_task
        def concat_then_split(wf_params, a, b):
            txt = ""
            with a as chunks:
                for chunk in chunks:
                    txt += chunk.read()

            out = Types.MultiPartBlob()  # Create at a random location specified in flytekit configuration
            with out.create_part('000000') as writer:
                writer.write("Chunk1")
            with out.create_part('000001') as writer:
                writer.write("Chunk2")
            b.set(out)
    """

    MultiPartCSV = _blobs.MultiPartCSV
    """
    See :py:attr:`flytekit.sdk.types.Types.MultiPartBlob`, but in CSV format
    """

    Schema = staticmethod(_schema.schema_instantiator)
    """
    Use this to specify a Schema blob object which is essentially a chunked stream of Parquet dataframes.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        Cast behavior:
            1) A generic schema (specified as `Types.Schema()`) can receive input from any schema type regardless of
               column definitions.
            2) A schema can receive as input any schema object as long as the upstream schema has a superset of the
               column names defined and the types match for paired columns.  Ordering does not matter.

        As input:
            1) If set, a :py:class:`flytekit.common.types.impl.schema.Schema` object will be received.
            2) If not set, a None value.

        As output:
            1) A user may specify a path string to a chunked dataframe non-recursive directory.
            2) A user may construct a :py:class:`flytekit.common.types.impl.schema.Schema` object (with the correct
               column definitions) and pass it as output.
            3) Output can be nulled with a None value.

        From command-line:
            Specify a path to the schema object. This path must be accessible from the container when
            executing--either by being downloaded from an accessible remote location like s3 or as a local file.

    .. code-block:: python

        @inputs(generic=Types.Schema(), typed=Types.Schema([('a', Types.Integer), ('b', Types.Float)]))
        @outputs(b=Types.Schema([('a', Types.Integer), ('b', Types.Float)]))
        @python_task
        def concat_then_split(wf_params, generic, typed,):
            with typed as reader:
                # Each chunk is loaded as a pandas.DataFrame object
                for df in reader.iter_chunks():
                    # Operate on the dataframe

            # Create at a random location specified in flytekit configuration
            out = Types.Schema([('a', Types.Integer), ('b', Types.Float)])()
            with out as writer:
                writer.write(
                    pandas.DataFrame.from_dict(
                        {
                            'a': [1, 2, 3],
                            'b': [5.0, 6.0, 7.0]
                        }
                    )
                )
            b.set(out)
    """

    Proto = staticmethod(_proto.create_protobuf)
    """
    Proto type wraps a protobuf type to provide interoperability between protobuf and flyte typing system. Using this
    type, you can define custom input/output variable types of flyte entities and continue to provide strong typing
    syntax. Proto type serializes proto objects as binary (leveraging `flyteidl's Binary literal <https://github.com/lyft/flyteidl/blob/793b09d190148236f41ad8160b5cec9a3325c16f/protos/flyteidl/core/literals.proto#L45>`_).
    Binary serialization of protobufs is the most space-efficient serialization form. Because of the way protobufs are
    designed, unmarshalling the serialized proto requires access to the corresponding type. In order to use/visualize
    the serialized proto, you will generally need to write custom code in the corresponding component.

    .. note::

        The protobuf Python library should be installed on the PYTHONPATH to ensure the type engine can access the
        appropriate Python code to deserialize the protobuf.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            1) If set, a Python protobuf object of the type specified in the definition.
            2) If not set, a None value.

        As output:
            1) A Python protobuf object matching the type specified by the users.
            2) Set None to null the output.

        From command-line:
            A base-64 encoded string of the serialized protobuf.

    .. code-block:: python

        from protos import my_protos_pb2

        @inputs(a=Types.Proto(my_protos_pb2.Custom))
        @outputs(b=Types.Proto(my_protos_pb2.Custom))
        @python_task
        def assert_and_create(wf_params, a, b):
            assert a.field1 == 1
            assert a.field2 == 'abc'
            b.set(
                my_protos_pb2.Custom(
                    field1=100,
                    field2='hello'
                )
            )
    """

    GenericProto = staticmethod(_proto.create_generic)
    """
    GenericProto type wraps a protobuf type to provide interoperability between protobuf and flyte typing system. Using
    this type, you can define custom input/output variable types of flyte entities and continue to provide strong typing
    syntax. Proto type serializes proto objects as binary (leveraging `flyteidl's Binary literal <https://github.com/lyft/flyteidl/blob/793b09d190148236f41ad8160b5cec9a3325c16f/protos/flyteidl/core/literals.proto#L63>`_).
    A generic proto is a specialization of the Generic type with added convenience functions to support marshalling/
    unmarshalling of the underlying protobuf object using the protobuf official json marshaller. While GenericProto type
    does not produce the most space-efficient representation of protobufs, it's a suitable solution for making protobufs
    easily accessible (i.e. humanly readable) in other flyte components (e.g. console, cli... etc.).

    .. note::

        The protobuf Python library should be installed on the PYTHONPATH to ensure the type engine can access the
        appropriate Python code to deserialize the protobuf.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            1) If set, a Python protobuf object of the type specified in the definition.
            2) If not set, a None value.

        As output:
            1) A Python protobuf object matching the type specified by the users.
            2) Set None to null the output.

        From command-line:
            A base-64 encoded string of the serialized protobuf.

    .. code-block:: python

        from protos import my_protos_pb2

        @inputs(a=Types.GenericProto(my_protos_pb2.Custom))
        @outputs(b=Types.GenericProto(my_protos_pb2.Custom))
        @python_task
        def assert_and_create(wf_params, a, b):
            assert a.field1 == 1
            assert a.field2 == 'abc'
            b.set(
                my_protos_pb2.Custom(
                    field1=100,
                    field2='hello'
                )
            )
    """

    List = staticmethod(_containers.List)
    """
    Use this to specify a list of any type--including nested lists.

    When used with an SDK-decorated method, expect this behavior from the default type engine:

        As input:
            1) If set, a Python list populated with values matching the behavior of the list's sub-type.
            2) If not set, a None value.

        As output:
            1) A Python list containing values adhering to the list's sub-type.
            2) Set None to null the output.

        From command-line:
            Specify a valid JSON list string.  The sub-values will be checked against the sub-type of the list.

    .. note::

        Shorthand syntax is supported of the form: `[Types.Integer]` in addition to longhand syntax like
        `Types.List(Types.Integer)`.  Both forms are equivalent.

    .. note::

        Lists can be arbitrarily deeply nested, however, the typing must be consistent between all sibling values in a
        nested list.  Syntax for nesting is `[[[Types.Integer]]]` to create a 3D list of integers.

    .. code-block:: python

        @inputs(a=[Types.Integer])
        @outputs(b=[Types.Integer])
        @python_task
        def square_each(wf_params, a, b):
            b.set(
                [x * x for x in a]
            )
    """



import abc as _abc
import datetime as _datetime
import logging as _logging
import sys as _sys
import time as _time
import traceback as _traceback

import six as _six


class Sensor(object, metaclass=_abc.ABCMeta):
    def __init__(self, evaluation_interval=None, max_failures=0):
        """
        :param datetime.timedelta evaluation_interval: This is the time to wait between evaluation attempts of this
            sensor. If the sensor takes longer to evaluate than the poll_interval, it will immediately begin
            evaluation again.
        :param int max_failures: This is the maximum number of failures that can happen while attempting to sense
            before perma-failing.
        """
        if evaluation_interval is None:
            evaluation_interval = _datetime.timedelta(seconds=30)
        self._evaluation_interval = evaluation_interval
        self._max_failures = max_failures
        self._failures = 0
        self._exc_info = None
        self._last_executed_time = _datetime.datetime(year=1990, month=6, day=30)  # Arbitrary date in the past.
        self._sensed = False

    @_abc.abstractmethod
    def _do_poll(self):
        """
        :rtype: (bool, Optional[datetime.timedelta])
        """
        pass

    def sense_with_wait_hint(self):
        """
        Attempts to sense based on the lambda expression.  The method will return the last sensed result.  If the
        rate of sensing is exceeded for the sensor, the timedelta in the returned tuple will tell the caller how long it
        should sleep before trying again.
        :rtype: (bool, Optional[datetime.timedelta])
        """
        # Return cached success.  This simplifies code for the conditional sensors.
        if self._sensed:
            return self._sensed, self._evaluation_interval

        # Perma-fail to prevent abuse of sensed objects.
        if self._failures > self._max_failures:
            _six.reraise(*self._exc_info)

        now = _datetime.datetime.utcnow()

        time_to_wait_eval_period = self._evaluation_interval - (now - self._last_executed_time)
        if time_to_wait_eval_period > _datetime.timedelta():
            return self._sensed, time_to_wait_eval_period

        try:
            self._sensed, time_to_wait = self._do_poll()
            time_to_wait = time_to_wait or self._evaluation_interval
        except BaseException:
            self._failures += 1
            self._exc_info = _sys.exc_info()
            if self._failures > self._max_failures:
                _logging.error(
                    "{} failed (with no remaining retries) due to:\n\n{}".format(self, _traceback.format_exc()),
                )
                raise
            else:
                _logging.warn("{} failed (but will retry) due to:\n\n{}".format(self, _traceback.format_exc()))
            time_to_wait = self._evaluation_interval

        self._last_executed_time = _datetime.datetime.utcnow()

        return self._sensed, time_to_wait

    def sense(self, timeout=None):
        """
        Attempts
        :param datetime.timedelta timeout:
        :rtype: bool
        """
        started = _datetime.datetime.utcnow()
        while True:
            sensed, time_to_wait = self.sense_with_wait_hint()
            if sensed:
                return True
            if time_to_wait:
                _time.sleep(time_to_wait.total_seconds())
            if timeout is not None and (_datetime.datetime.utcnow() - started) > timeout:
                return False



import flyteidl.admin.execution_pb2 as _execution_pb2
import flyteidl.admin.node_execution_pb2 as _node_execution_pb2
import flyteidl.admin.task_execution_pb2 as _task_execution_pb2
import pytz as _pytz

from flytekit.models import common as _common_models
from flytekit.models import literals as _literals_models
from flytekit.models.core import execution as _core_execution
from flytekit.models.core import identifier as _identifier


class ExecutionMetadata(_common_models.FlyteIdlEntity):
    class ExecutionMode(object):
        MANUAL = 0
        SCHEDULED = 1
        SYSTEM = 2

    def __init__(self, mode, principal, nesting):
        """
        :param int mode: An enum value from ExecutionMetadata.ExecutionMode which specifies how the job started.
        :param Text principal: The entity that triggered the execution
        :param int nesting: An integer representing how deeply nested the workflow is (i.e. was it triggered by a parent
            workflow)
        """
        self._mode = mode
        self._principal = principal
        self._nesting = nesting

    @property
    def mode(self):
        """
        An enum value from ExecutionMetadata.ExecutionMode which specifies how the job started.
        :rtype: int
        """
        return self._mode

    @property
    def principal(self):
        """
        The entity that triggered the execution
        :rtype: Text
        """
        return self._principal

    @property
    def nesting(self):
        """
        An integer representing how deeply nested the workflow is (i.e. was it triggered by a parent workflow)
        :rtype: int
        """
        return self._nesting

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.execution_pb2.ExecutionMetadata
        """
        return _execution_pb2.ExecutionMetadata(mode=self.mode, principal=self.principal, nesting=self.nesting)

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.execution_pb2.ExecutionMetadata pb2_object:
        :return: ExecutionMetadata
        """
        return cls(
            mode=pb2_object.mode,
            principal=pb2_object.principal,
            nesting=pb2_object.nesting,
        )


class ExecutionSpec(_common_models.FlyteIdlEntity):
    def __init__(
        self,
        launch_plan,
        metadata,
        notifications=None,
        disable_all=None,
        labels=None,
        annotations=None,
        auth_role=None,
    ):
        """
        :param flytekit.models.core.identifier.Identifier launch_plan: Launch plan unique identifier to execute
        :param ExecutionMetadata metadata: The metadata to be associated with this execution
        :param NotificationList notifications: List of notifications for this execution.
        :param bool disable_all: If true, all notifications should be disabled.
        :param flytekit.models.common.Labels labels: Labels to apply to the execution.
        :param flytekit.models.common.Annotations annotations: Annotations to apply to the execution
        :param flytekit.models.common.AuthRole auth_role: The authorization method with which to execute the workflow.

        """
        self._launch_plan = launch_plan
        self._metadata = metadata
        self._notifications = notifications
        self._disable_all = disable_all
        self._labels = labels or _common_models.Labels({})
        self._annotations = annotations or _common_models.Annotations({})
        self._auth_role = auth_role or _common_models.AuthRole()

    @property
    def launch_plan(self):
        """
        If the values were too large, this is the URI where the values were offloaded.
        :rtype: flytekit.models.core.identifier.Identifier
        """
        return self._launch_plan

    @property
    def metadata(self):
        """
        :rtype: ExecutionMetadata
        """
        return self._metadata

    @property
    def notifications(self):
        """
        :rtype: Optional[NotificationList]
        """
        return self._notifications

    @property
    def disable_all(self):
        """
        :rtype: Optional[bool]
        """
        return self._disable_all

    @property
    def labels(self):
        """
        :rtype: flytekit.models.common.Labels
        """
        return self._labels

    @property
    def annotations(self):
        """
        :rtype: flytekit.models.common.Annotations
        """
        return self._annotations

    @property
    def auth_role(self):
        """
        :rtype: flytekit.models.common.AuthRole
        """
        return self._auth_role

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.execution_pb2.ExecutionSpec
        """
        return _execution_pb2.ExecutionSpec(
            launch_plan=self.launch_plan.to_flyte_idl(),
            metadata=self.metadata.to_flyte_idl(),
            notifications=self.notifications.to_flyte_idl() if self.notifications else None,
            disable_all=self.disable_all,
            labels=self.labels.to_flyte_idl(),
            annotations=self.annotations.to_flyte_idl(),
            auth_role=self._auth_role.to_flyte_idl() if self.auth_role else None,
        )

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.admin.execution_pb2.ExecutionSpec p:
        :return: ExecutionSpec
        """
        return cls(
            launch_plan=_identifier.Identifier.from_flyte_idl(p.launch_plan),
            metadata=ExecutionMetadata.from_flyte_idl(p.metadata),
            notifications=NotificationList.from_flyte_idl(p.notifications) if p.HasField("notifications") else None,
            disable_all=p.disable_all if p.HasField("disable_all") else None,
            labels=_common_models.Labels.from_flyte_idl(p.labels),
            annotations=_common_models.Annotations.from_flyte_idl(p.annotations),
            auth_role=_common_models.AuthRole.from_flyte_idl(p.auth_role),
        )


class LiteralMapBlob(_common_models.FlyteIdlEntity):
    def __init__(self, values=None, uri=None):
        """
        :param flytekit.models.literals.LiteralMap values:
        :param Text uri:
        """
        self._values = values
        self._uri = uri

    @property
    def values(self):
        """
        :rtype: flytekit.models.literals.LiteralMap
        """
        return self._values

    @property
    def uri(self):
        """
        :rtype: Text
        """
        return self._uri

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.execution_pb2.LiteralMapBlob
        """
        return _execution_pb2.LiteralMapBlob(
            values=self.values.to_flyte_idl() if self.values is not None else None,
            uri=self.uri,
        )

    @classmethod
    def from_flyte_idl(cls, pb):
        """
        :param flyteidl.admin.execution_pb2.LiteralMapBlob pb:
        :rtype: LiteralMapBlob
        """
        values = None
        if pb.HasField("values"):
            values = LiteralMapBlob.from_flyte_idl(pb.values)
        return cls(values=values, uri=pb.uri if pb.HasField("uri") else None)


class Execution(_common_models.FlyteIdlEntity):
    def __init__(self, id, spec, closure):
        """
        :param flytekit.models.core.identifier.WorkflowExecutionIdentifier id:
        :param Text id:
        :param ExecutionSpec spec:
        :param ExecutionClosure closure:
        """
        self._id = id
        self._spec = spec
        self._closure = closure

    @property
    def id(self):
        """
        :rtype: flytekit.models.core.identifier.WorkflowExecutionIdentifier
        """
        return self._id

    @property
    def closure(self):
        """
        :rtype: ExecutionClosure
        """
        return self._closure

    @property
    def spec(self):
        """
        :rtype: ExecutionSpec
        """
        return self._spec

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.execution_pb2.Execution
        """
        return _execution_pb2.Execution(
            id=self.id.to_flyte_idl(),
            closure=self.closure.to_flyte_idl(),
            spec=self.spec.to_flyte_idl(),
        )

    @classmethod
    def from_flyte_idl(cls, pb):
        """
        :param flyteidl.admin.execution_pb2.Execution pb:
        :rtype: Execution
        """
        return cls(
            id=_identifier.WorkflowExecutionIdentifier.from_flyte_idl(pb.id),
            closure=ExecutionClosure.from_flyte_idl(pb.closure),
            spec=ExecutionSpec.from_flyte_idl(pb.spec),
        )


class ExecutionClosure(_common_models.FlyteIdlEntity):
    def __init__(self, phase, started_at, error=None, outputs=None):
        """
        :param int phase: From the flytekit.models.core.execution.WorkflowExecutionPhase enum
        :param datetime.datetime started_at:
        :param flytekit.models.core.execution.ExecutionError error:
        :param LiteralMapBlob outputs:
        """
        self._phase = phase
        self._started_at = started_at
        self._error = error
        self._outputs = outputs

    @property
    def error(self):
        """
        :rtype: flytekit.models.core.execution.ExecutionError
        """
        return self._error

    @property
    def phase(self):
        """
        From the flytekit.models.core.execution.WorkflowExecutionPhase enum
        :rtype: int
        """
        return self._phase

    @property
    def started_at(self):
        """
        :rtype: datetime.datetime
        """
        return self._started_at

    @property
    def outputs(self):
        """
        :rtype: LiteralMapBlob
        """
        return self._outputs

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.execution_pb2.ExecutionClosure
        """
        obj = _execution_pb2.ExecutionClosure(
            phase=self.phase,
            error=self.error.to_flyte_idl() if self.error is not None else None,
            outputs=self.outputs.to_flyte_idl() if self.outputs is not None else None,
        )
        obj.started_at.FromDatetime(self.started_at.astimezone(_pytz.UTC).replace(tzinfo=None))
        return obj

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.execution_pb2.ExecutionClosure pb2_object:
        :rtype: ExecutionClosure
        """
        error = None
        if pb2_object.HasField("error"):
            error = _core_execution.ExecutionError.from_flyte_idl(pb2_object.error)
        outputs = None
        if pb2_object.HasField("outputs"):
            outputs = LiteralMapBlob.from_flyte_idl(pb2_object.outputs)
        return cls(
            error=error,
            outputs=outputs,
            phase=pb2_object.phase,
            started_at=pb2_object.started_at.ToDatetime().replace(tzinfo=_pytz.UTC),
        )


class NotificationList(_common_models.FlyteIdlEntity):
    def __init__(self, notifications):
        """
        :param list[flytekit.models.common.Notification] notifications: A simple list of notifications.
        """
        self._notifications = notifications

    @property
    def notifications(self):
        """
        :rtype: list[flytekit.models.common.Notification]
        """
        return self._notifications

    def to_flyte_idl(self):
        """
        :rtype:  flyteidl.admin.execution_pb2.NotificationList
        """
        return _execution_pb2.NotificationList(notifications=[n.to_flyte_idl() for n in self.notifications])

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.execution_pb2.NotificationList pb2_object:
        :rtype: NotificationList
        """
        return cls([_common_models.Notification.from_flyte_idl(p) for p in pb2_object.notifications])


class _CommonDataResponse(_common_models.FlyteIdlEntity):
    """
    Currently, node, task, and workflow execution all have the same get data response. So we'll create this common
    superclass to reduce code duplication until things diverge in the future.
    """

    def __init__(self, inputs, outputs, full_inputs, full_outputs):
        """
        :param _common_models.UrlBlob inputs:
        :param _common_models.UrlBlob outputs:
        :param _literals_pb2.LiteralMap full_inputs:
        :param _literals_pb2.LiteralMap full_outputs:
        """
        self._inputs = inputs
        self._outputs = outputs
        self._full_inputs = full_inputs
        self._full_outputs = full_outputs

    @property
    def inputs(self):
        """
        :rtype: _common_models.UrlBlob
        """
        return self._inputs

    @property
    def outputs(self):
        """
        :rtype: _common_models.UrlBlob
        """
        return self._outputs

    @property
    def full_inputs(self):
        """
        :rtype: _literals_pb2.LiteralMap
        """
        return self._full_inputs

    @property
    def full_outputs(self):
        """
        :rtype: _literals_pb2.LiteralMap
        """
        return self._full_outputs


class WorkflowExecutionGetDataResponse(_CommonDataResponse):
    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param _execution_pb2.WorkflowExecutionGetDataResponse pb2_object:
        :rtype: WorkflowExecutionGetDataResponse
        """
        return cls(
            inputs=_common_models.UrlBlob.from_flyte_idl(pb2_object.inputs),
            outputs=_common_models.UrlBlob.from_flyte_idl(pb2_object.outputs),
            full_inputs=_literals_models.LiteralMap.from_flyte_idl(pb2_object.full_inputs),
            full_outputs=_literals_models.LiteralMap.from_flyte_idl(pb2_object.full_outputs),
        )

    def to_flyte_idl(self):
        """
        :rtype: _execution_pb2.WorkflowExecutionGetDataResponse
        """
        return _execution_pb2.WorkflowExecutionGetDataResponse(
            inputs=self.inputs.to_flyte_idl(),
            outputs=self.outputs.to_flyte_idl(),
            full_inputs=self.full_inputs.to_flyte_idl(),
            full_outputs=self.full_outputs.to_flyte_idl(),
        )


class TaskExecutionGetDataResponse(_CommonDataResponse):
    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param _task_execution_pb2.TaskExecutionGetDataResponse pb2_object:
        :rtype: TaskExecutionGetDataResponse
        """
        return cls(
            inputs=_common_models.UrlBlob.from_flyte_idl(pb2_object.inputs),
            outputs=_common_models.UrlBlob.from_flyte_idl(pb2_object.outputs),
            full_inputs=_literals_models.LiteralMap.from_flyte_idl(pb2_object.full_inputs),
            full_outputs=_literals_models.LiteralMap.from_flyte_idl(pb2_object.full_outputs),
        )

    def to_flyte_idl(self):
        """
        :rtype: _task_execution_pb2.TaskExecutionGetDataResponse
        """
        return _task_execution_pb2.TaskExecutionGetDataResponse(
            inputs=self.inputs.to_flyte_idl(),
            outputs=self.outputs.to_flyte_idl(),
            full_inputs=self.full_inputs.to_flyte_idl(),
            full_outputs=self.full_outputs.to_flyte_idl(),
        )


class NodeExecutionGetDataResponse(_CommonDataResponse):
    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param _node_execution_pb2.NodeExecutionGetDataResponse pb2_object:
        :rtype: NodeExecutionGetDataResponse
        """
        return cls(
            inputs=_common_models.UrlBlob.from_flyte_idl(pb2_object.inputs),
            outputs=_common_models.UrlBlob.from_flyte_idl(pb2_object.outputs),
            full_inputs=_literals_models.LiteralMap.from_flyte_idl(pb2_object.full_inputs),
            full_outputs=_literals_models.LiteralMap.from_flyte_idl(pb2_object.full_outputs),
        )

    def to_flyte_idl(self):
        """
        :rtype: _node_execution_pb2.NodeExecutionGetDataResponse
        """
        return _node_execution_pb2.NodeExecutionGetDataResponse(
            inputs=self.inputs.to_flyte_idl(),
            outputs=self.outputs.to_flyte_idl(),
            full_inputs=self.full_inputs.to_flyte_idl(),
            full_outputs=self.full_outputs.to_flyte_idl(),
        )
